# -*- coding: utf-8 -*-


"""Molecular constitutional and topological indices."""

from rdkit import Chem
from rdkit.Chem import Lipinski as LPK


class Constitution:
    """Constitutional descriptors."""

    @staticmethod
    def calculate_heavy_mol_weight(mol: Chem.Mol) -> float:
        """calculate_ molecular weight of heavy atoms."""
        MolWeight = 0
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() != 1:
                MolWeight += atom.GetMass()
        return MolWeight

    @staticmethod
    def calculate_average_mol_weight(mol: Chem.Mol) -> float:
        """calculate_ average molecular weight of heavy atoms."""
        MolWeight = 0
        for atom in mol.GetAtoms():
            MolWeight += atom.GetMass()
        return MolWeight / mol.GetNumAtoms()

    @staticmethod
    def calculate_hydrogen_number(mol: Chem.Mol) -> float:
        """calculate_ number of Hydrogens."""
        i = 0
        Hmol = Chem.AddHs(mol)
        for atom in Hmol.GetAtoms():
            if atom.GetAtomicNum() == 1:
                i += 1
        return i

    @staticmethod
    def calculate_halogen_number(mol: Chem.Mol) -> float:
        """calculate_ number of Halogens."""
        i = 0
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() in [9, 17, 35, 53]:
                i += 1
        return i

    @staticmethod
    def calculate_hetero_number(mol: Chem.Mol) -> float:
        """calculate_ number of Heteroatoms."""
        i = 0
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() not in [1, 6]:
                i += 1
        return mol.GetNumAtoms() - i

    @staticmethod
    def calculate_heavy_atom_number(mol: Chem.Mol) -> float:
        """calculate_ number of Heavy atoms."""
        return mol.GetNumHeavyAtoms()

    @staticmethod
    def _calculate_element_number(mol: Chem.Mol, Atomic_number=6) -> float:
        """calculate_ number of atoms with specified element."""
        i = 0
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() == Atomic_number:
                i += 1
        return i

    @staticmethod
    def calculate_fluorine_number(mol: Chem.Mol) -> float:
        """calculate_ number of Fluorine atoms."""
        return Constitution._calculate_element_number(mol, Atomic_number=9)

    @staticmethod
    def calculate_chlorine_number(mol: Chem.Mol) -> float:
        """calculate_ number of Fluorine atoms."""
        return Constitution._calculate_element_number(mol, Atomic_number=17)


    def calculate_bromine_number(mol: Chem.Mol) -> float:
        """calculate_ number of Bromine atoms."""
        return Constitution._calculate_element_number(mol, Atomic_number=35)

    @staticmethod
    def calculate_iodine_number(mol: Chem.Mol) -> float:
        """calculate_ number of Iodine atoms."""
        return Constitution._calculate_element_number(mol, Atomic_number=53)

    @staticmethod
    def calculate_carbon_number(mol: Chem.Mol) -> float:
        """calculate_ number of Carbon atoms."""
        return Constitution._calculate_element_number(mol, Atomic_number=6)

    @staticmethod
    def calculate_phosphor_number(mol: Chem.Mol) -> float:
        """Calcualtion number of Phosphor atoms."""
        return Constitution._calculate_element_number(mol, Atomic_number=15)

    @staticmethod
    def calculate_sulfur_number(mol: Chem.Mol) -> float:
        """calculate_ number of Sulfur atoms."""
        return Constitution._calculate_element_number(mol, Atomic_number=16)

    @staticmethod
    def calculate_oxygen_number(mol: Chem.Mol) -> float:
        """calculate_ number of Oxygen atoms."""
        return Constitution._calculate_element_number(mol, Atomic_number=8)

    @staticmethod
    def calculate_nitrogen_number(mol: Chem.Mol) -> float:
        """calculate_ number of Nitrogen atoms."""
        return Constitution._calculate_element_number(mol, Atomic_number=7)

    @staticmethod
    def calculate_ring_number(mol: Chem.Mol) -> float:
        """calculate_ number of rings."""
        return len(Chem.GetSSSR(mol))

    @staticmethod
    def calculate_rot_bond_number(mol: Chem.Mol) -> float:
        """calculate_ number of rotatable bonds."""
        return LPK.NumRotatableBonds(mol)

    @staticmethod
    def calculate_hdonor_number(mol: Chem.Mol) -> float:
        """calculate_ number of Hydrongen bond donors."""
        return LPK.NumHDonors(mol)

    @staticmethod
    def calculate_hacceptor_number(mol: Chem.Mol) -> float:
        """calculate_ number of Hydrogen bond acceptors."""
        return LPK.NumHAcceptors(mol)

    @staticmethod
    def calculate_singlebond_number(mol: Chem.Mol) -> float:
        """calculate_ number of single bonds."""
        i = 0
        for bond in mol.GetBonds():
            if bond.GetBondType().name == 'SINGLE':
                i += 1
        return i

    @staticmethod
    def calculate_doublebond_number(mol: Chem.Mol) -> float:
        """calculate_ number of double bonds."""
        i = 0
        for bond in mol.GetBonds():
            if bond.GetBondType().name == 'DOUBLE':
                i += 1
        return i

    @staticmethod
    def calculate_triplebond_number(mol: Chem.Mol) -> float:
        """calculate_ number of triple bonds."""
        i = 0
        for bond in mol.GetBonds():
            if bond.GetBondType().name == 'TRIPLE':
                i += 1
        return i

    @staticmethod
    def calculate_aromaticbond_number(mol: Chem.Mol) -> float:
        """calculate_ number of aromatic bonds."""
        i = 0
        for bond in mol.GetBonds():
            if bond.GetBondType().name == 'AROMATIC':
                i += 1
        return i

    @staticmethod
    def calculate_allatom_number(mol: Chem.Mol) -> float:
        """calculate_ number of all atoms."""
        return Chem.AddHs(mol).GetNumAtoms()

    @staticmethod
    def _calculate_path_n(mol: Chem.Mol, path_Length=2) -> float:
        """calculate_ number of path of length N."""
        return len(Chem.FindAllPathsOfLengthN(mol, path_Length, useBonds=1))

    @staticmethod
    def calculate_path_1(mol: Chem.Mol) -> float:
        """calculate_ number of path length of 1."""
        return Constitution._calculate_path_n(mol, 1)

    @staticmethod
    def calculate_path_2(mol: Chem.Mol) -> float:
        """calculate_ number of path length of 2."""
        return Constitution._calculate_path_n(mol, 2)

    @staticmethod
    def calculate_path_3(mol: Chem.Mol) -> float:
        """calculate_ number of path length of 3."""
        return Constitution._calculate_path_n(mol, 3)

    @staticmethod
    def calculate_path_4(mol: Chem.Mol) -> float:
        """calculate_ number of path length of 4."""
        return Constitution._calculate_path_n(mol, 4)

    @staticmethod
    def calculate_path_5(mol: Chem.Mol) -> float:
        """calculate_ number of path length of 5."""
        return Constitution._calculate_path_n(mol, 5)

    @staticmethod
    def calculate_path_6(mol: Chem.Mol) -> float:
        """calculate_ number of path length of 6."""
        return Constitution._calculate_path_n(mol, 6)


    _constitutional = {'Weight': calculate_heavy_mol_weight,
                       'nH': calculate_hydrogen_number,
                       'nHal': calculate_halogen_number,
                       'nHet': calculate_hetero_number,
                       'nHA': calculate_heavy_atom_number,
                       'nF': calculate_fluorine_number,
                       'nCl': calculate_chlorine_number,
                       'nBr': calculate_bromine_number,
                       'nI': calculate_iodine_number,
                       'nC': calculate_carbon_number,
                       'nP': calculate_phosphor_number,
                       'nS': calculate_oxygen_number,
                       'nO': calculate_oxygen_number,
                       'nN': calculate_nitrogen_number,
                       'nRing': calculate_ring_number,
                       'nRotB': calculate_rot_bond_number,
                       'nHBD': calculate_hdonor_number,
                       'nHBA': calculate_hacceptor_number,
                       'nSBond': calculate_singlebond_number,
                       'nDBond': calculate_doublebond_number,
                       'nTBond': calculate_triplebond_number,
                       'nAroBond': calculate_aromaticbond_number,
                       'nAtom': calculate_allatom_number,
                       'AWeight': calculate_average_mol_weight,
                       'path_L1': calculate_path_1,
                       'path_L2': calculate_path_2,
                       'path_L3': calculate_path_3,
                       'path_L4': calculate_path_4,
                       'path_L5': calculate_path_5,
                       'path_L6': calculate_path_6,
                       }

    @staticmethod
    def get_all(mol: Chem.Mol) -> dict:
        """Get all (30) constitutional descriptors."""
        result = {}
        for DesLabel in Constitution._constitutional.keys():
            result[DesLabel] = Constitution._constitutional[DesLabel](mol)
        return result




# -*- coding: utf-8 -*-


"""Kier and Hall's kappa indices."""


from rdkit import Chem
from rdkit.Chem import rdchem

periodicTable = rdchem.GetPeriodicTable()

class Kappa:
    """Kier and Hall's kappa indices."""

    @staticmethod
    def calculate_kappa1(mol: Chem.Mol) -> float:
        """Calculate molecular shape index for one bonded fragment."""
        P1 = mol.GetNumBonds(onlyHeavy=1)
        A = mol.GetNumHeavyAtoms()
        denom = P1 + 0.0
        if denom:
            kappa = (A) * (A - 1) ** 2 / denom ** 2
        else:
            kappa = 0.0
        return kappa

    @staticmethod
    def calculate_kappa2(mol: Chem.Mol) -> float:
        """Calculate molecular shape index for two bonded fragments."""
        P2 = len(Chem.FindAllPathsOfLengthN(mol, 2))
        A = mol.GetNumHeavyAtoms()
        denom = P2 + 0.0
        if denom:
            kappa = (A - 1) * (A - 2) ** 2 / denom ** 2
        else:
            kappa = 0.0
        return kappa

    @staticmethod
    def calculate_kappa3(mol: Chem.Mol) -> float:
        """Calculate molecular shape index for three bonded fragments."""
        P3 = len(Chem.FindAllPathsOfLengthN(mol, 3))
        A = mol.GetNumHeavyAtoms()
        denom = P3 + 0.0
        if denom:
            if A % 2 == 1:
                kappa = (A - 1) * (A - 3) ** 2 / denom ** 2
            else:
                kappa = (A - 3) * (A - 2) ** 2 / denom ** 2
        else:
            kappa = 0.0
        return kappa

    hall_kier_alphas = {'Br': [None, None, 0.48],
                      'C': [-0.22, -0.13, 0.0],
                      'Cl': [None, None, 0.29],
                      'F': [None, None, -0.07],
                      'H': [0.0, 0.0, 0.0],
                      'I': [None, None, 0.73],
                      'N': [-0.29, -0.2, -0.04],
                      'O': [None, -0.2, -0.04],
                      'P': [None, 0.3, 0.43],
                      'S': [None, 0.22, 0.35]}

    @staticmethod
    def _hall_kier_alpha(mol: Chem.Mol) -> float:
        """Calculate Hall-Kier alpha value for a molecule."""
        alphaSum = 0.0
        rC = periodicTable.GetRb0(6)
        for atom in mol.GetAtoms():
            atNum = atom.GetAtomicNum()
            if not atNum:
                continue
            symb = atom.GetSymbol()
            alphaV = Kappa.hall_kier_alphas.get(symb, None)
            if alphaV is not None:
                hyb = atom.GetHybridization() - 2
                if hyb < len(alphaV):
                    alpha = alphaV[hyb]
                    if alpha is None:
                        alpha = alphaV[-1]
                else:
                    alpha = alphaV[-1]
            else:
                rA = periodicTable.GetRb0(atNum)
                alpha = rA / rC - 1
            alphaSum += alpha
        return alphaSum

    @staticmethod
    def calculate_kappa_alapha1(mol: Chem.Mol) -> float:
        """Calculate molecular shape index for one bonded fragment."""
        P1 = mol.GetNumBonds(onlyHeavy=1)
        A = mol.GetNumHeavyAtoms()
        alpha = Kappa._hall_kier_alpha(mol)
        denom = P1 + alpha
        if denom:
            kappa = (A + alpha) * (A + alpha - 1) ** 2 / denom ** 2
        else:
            kappa = 0.0
        return kappa

    @staticmethod
    def calculate_kappa_alapha2(mol: Chem.Mol) -> float:
        """Calculate molecular shape index for two bonded fragments."""
        P2 = len(Chem.FindAllPathsOfLengthN(mol, 2))
        A = mol.GetNumHeavyAtoms()
        alpha = Kappa._hall_kier_alpha(mol)
        denom = P2 + alpha
        if denom:
            kappa = (A + alpha - 1) * (A + alpha - 2) ** 2 / denom ** 2
        else:
            kappa = 0.0
        return kappa

    @staticmethod
    def calculate_kappa_alapha3(mol: Chem.Mol) -> float:
        """Calculate molecular shape index for three bonded fragments."""
        P3 = len(Chem.FindAllPathsOfLengthN(mol, 3))
        A = mol.GetNumHeavyAtoms()
        alpha = Kappa._hall_kier_alpha(mol)
        denom = P3 + alpha
        if denom:
            if A % 2 == 1:
                kappa = (A + alpha - 1) * (A + alpha - 3) ** 2 / denom ** 2
            else:
                kappa = (A + alpha - 3) * (A + alpha - 2) ** 2 / denom ** 2
        else:
            kappa = 0.0
        return kappa

    @staticmethod
    def calculate_flexibility(mol: Chem.Mol) -> float:
        """Calculate Kier molecular flexibility index."""
        kappa1 = Kappa.calculate_kappa_alapha1(mol)
        kappa2 = Kappa.calculate_kappa_alapha2(mol)
        A = mol.GetNumHeavyAtoms()
        phi = kappa1 * kappa2 / (A + 0.0)
        return phi

    @staticmethod
    def get_all(mol: Chem.Mol) -> dict:
        """Calculate all (7) kappa values."""
        res = {}
        res['kappam1'] = Kappa.calculate_kappa_alapha1(mol)
        res['kappam2'] = Kappa.calculate_kappa_alapha2(mol)
        res['kappam3'] = Kappa.calculate_kappa_alapha3(mol)
        res['phi'] = Kappa.calculate_flexibility(mol)
        res['kappa1'] = Kappa.calculate_kappa1(mol)
        res['kappa2'] = Kappa.calculate_kappa2(mol)
        res['kappa3'] = Kappa.calculate_kappa3(mol)
        return res



# -*- coding: utf-8 -*-

"""Tests for the fetching of molecules from online resources."""

import unittest

from chemopy import MolFrom

from tests.constants import *


class TestDescriptorsAndFingerprints(unittest.TestCase):
    """Tests for online molecule retrieval."""

    def test_cas(self):
        for casnum in CAS_NUMBERS:
            mol = MolFrom.CAS(casnum)
            self.assertIsNotNone(mol)
            self.assertNotEqual(Chem.MolToSmiles(mol), '')

    def test_ebi(self):
        for ebinum in EBI_NUMBERS:
            mol = MolFrom.EBI(ebinum)
            self.assertIsNotNone(mol)
            self.assertNotEqual(Chem.MolToSmiles(mol), '')

    def test_ncbi(self):
        for ncbinum in NCBI_NUMBERS:
            mol = MolFrom.NCBI(ncbinum)
            self.assertIsNotNone(mol)
            self.assertNotEqual(Chem.MolToSmiles(mol), '')

    def test_drugbank(self):
        for drugbankid in DRUGBANK_NUMBERS:
            mol = MolFrom.DrugBank(drugbankid)
            self.assertIsNotNone(mol)
            self.assertNotEqual(Chem.MolToSmiles(mol), '')

    def test_kegg(self):
        for keggid in KEGG_NUMBERS:
            mol = MolFrom.KEGG(keggid)
            self.assertIsNotNone(mol)
            self.assertNotEqual(Chem.MolToSmiles(mol), '')


# -*- coding: utf-8 -*-


"""Class to manipulate 3D vectors."""


from __future__ import annotations

import math
import random
from typing import Iterable

import numpy as np

RAD2DEG = 180.0 / math.pi
DEG2RAD = math.pi / 180.0
SMALL = 1E-6


def is_near_zero(x: float, epsilon: float = None) -> bool:
    """Return if x is lower than epsilon.

    :param x: small number to test
    :param epsilon: extremelly small number x needs to be smaller than
                    (default: 1e-6)
    """
    return x < SMALL if epsilon is None else x < epsilon


class Vector3D:
    """Class holding 3D vector coordinates."""

    def __init__(self, x: float = 0.0, y: float = 0.0, z: float = 0.0) -> None:
        """Initialize a Vector3D."""
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)

    def __add__(self, rhs: Vector3D) -> Vector3D:
        """Return the resulting vector from adding another vector to the current one.

        :param rhs: other Vector3D.
        """
        return Vector3D(rhs.x + self.x, rhs.y + self.y, rhs.z + self.z)

    def __sub__(self, rhs: Vector3D) -> Vector3D:
        """Return the resulting vector from substracting another vector from the current one.

        :param rhs: other Vector3D.
        """
        return Vector3D(self.x - rhs.x, self.y - rhs.y, self.z - rhs.z)

    def __mul__(self, rhs: Vector3D) -> Vector3D:
        """Return the Hadamard product of the two vectors."""
        return Vector3D(self.x * rhs.x, self.y * rhs.y, self.z * rhs.z)

    def __pow__(self, n: int) -> Vector3D:
        """Execute n times the Hadamard product."""
        temp, acc = self.copy(), self.copy()
        for _ in range(n):
            acc *= temp
        return acc

    def __neg__(self) -> Vector3D:
        """Return the opposite vector of the current one."""
        return Vector3D(-self.x, -self.y, -self.z)

    def __pos__(self) -> Vector3D:
        """Return a copy of the current vector."""
        return Vector3D(self.x, self.y, self.z)

    def __eq__(self, rhs: Vector3D) -> bool:
        """Test for close equality between two vectors.

        :param rhs: other Vector3D.
        """
        return (is_near_zero(self.x - rhs.x) and is_near_zero(self.y - rhs.y) and is_near_zero(self.z - rhs.z))

    def __ne__(self, rhs: Vector3D) -> bool:
        """Test for inequality between two vectors.

        :param rhs: other Vector3D.
        """
        return not (self == rhs)

    def __str__(self) -> str:
        """Pretty print representation."""
        return f'({self.x:.2f}, {self.y:.2f}, {self.z:.2f})'

    def __repr__(self) -> str:
        """Unambiguous representation."""
        return f'Vector3D({self.x:.2f}, {self.y:.2f}, {self.z:.2f})'

    def copy(self) -> Vector3D:
        """Copy the current vector."""
        return Vector3D(self.x, self.y, self.z)

    def dot(self, rhs: Vector3D) -> Vector3D:
        """Return the dot product of the vector with another one."""
        return self.x * rhs.x + self.y * rhs.y + self.z * rhs.z

    def cross(self, rhs: Vector3D) -> Vector3D:
        """Return the cross product of the vector with another one."""
        return Vector3D(self.y * rhs.z - self.z * rhs.y,
                        self.z * rhs.x - self.x * rhs.z,
                        self.x * rhs.y - self.y * rhs.x)

    def length_sq(self) -> float:
        """Return the square value of the euclidean norm."""
        return self.x * self.x + self.y * self.y + self.z * self.z

    def length(self) -> float:
        """Return the euclidean norm."""
        return math.sqrt(self.length_sq())

    def scale(self, scale: float) -> Vector3D:
        """Scale the vector by a scalar.

        :param scale: value to scale each coordinate by.
        """
        self.x *= scale
        self.y *= scale
        self.z *= scale

    def normalize(self) -> Vector3D:
        """Scale the vector to unit length."""
        self.scale(1.0 / self.length())

    def scaled_vec(self, scale: float) -> Vector3D:
        """Return a copy scaled to specified length.

        :param scale: value to scale each coordinate by.
        """
        v = self.copy()
        v.scale(scale)
        return v

    def normal_vec(self) -> Vector3D:
        """Return a copy scaled to unit length."""
        return self.scaled_vec(1.0 / self.length())

    def parallel_vec(self, axis: Vector3D) -> Vector3D:
        """Return a vector parallel to the current one using the specified axis.

        :param axis: axis along which to search for the parallel vector
        """
        axis_len = axis.length()
        if is_near_zero(axis_len):
            result = self
        else:
            result = axis.scaled_vec(self.dot(axis) / axis.length() / axis.length())
        return result

    def perpendicular_vec(self, axis) -> Vector3D:
        """Return a vector perpendicular to the current one using the specified axis.

        :param axis: axis along which to search for the perpendicular vector
        """
        return self - self.parallel_vec(axis)

    def transform(self, matrix) -> Vector3D:
        """Apply transformation from a matrix to the vector.

        :param matrix: transforamtion matrix
        """
        x = matrix.elem00 * self.x + \
            matrix.elem10 * self.y + \
            matrix.elem20 * self.z + \
            matrix.elem30
        y = matrix.elem01 * self.x + \
            matrix.elem11 * self.y + \
            matrix.elem21 * self.z + \
            matrix.elem31
        z = matrix.elem02 * self.x + \
            matrix.elem12 * self.y + \
            matrix.elem22 * self.z + \
            matrix.elem32
        self.x, self.y, self.z = x, y, z

    def diff_length_sq(self, rhs: Vector3D) -> float:
        """Return the squared euclidean norm of the difference of the two vectors.

        :param rhs: other Vector3D.
        """
        diff = self - rhs
        return diff.dot(diff)

    def diff_length(self, rhs: Vector3D) -> float:
        """Return the squared euclidean norm of the difference of the two vectors.

        :param rhs: other Vector3D.
        """
        return math.sqrt(self.diff_length_sq(rhs))

    def to_numpy_row(self):
        """Transform to a numpy matrix row."""
        return np.array([self.x, self.y, self.z], dtype=float)

    def to_numpy_col(self):
        """Transform to a numpy matrix column."""
        return self.to_numpy_row().T

    @staticmethod
    def from_numpy(array: np.array) -> Vector3D:
        """Instanciate a Vector3D from numpy array."""
        if len(array) > 3:
            raise ValueError(f'numpy.array({array}) has more values than Vector3D can hold.')
        return Vector3D(array[0], array[1], array[2])


def normalize_angle(angle: float):
    """Normalize angle to be between -pi and +pi.

    :param angle: angle in radians
    """
    while abs(angle) > math.pi:
        if angle > math.pi:
            angle -= math.pi * 2
        if angle < -math.pi:
            angle += 2 * math.pi
    if is_near_zero(abs(angle + math.pi)):
        angle = math.pi
    return angle


def angle_diff(angle1: float, angle2: float, radians: bool = True) -> float:
    """Calculate the difference between two angles.

    :param angle1: angle (in radians unless specified)
    :param angle2: angle (in radians unless specified)
    """
    if not radians:
        angle1 = angle1 * math.pi / 180
        angle2 = angle2 * math.pi / 180
    norm_angle1 = normalize_angle(angle1)
    norm_angle2 = normalize_angle(angle2)
    diff = normalize_angle(norm_angle1 - norm_angle2)
    return diff if radians else radians / math.pi * 180


def pos_distance(p1, p2):
    """Return the euclidean norm of the difference of the two vectors."""
    return math.sqrt(pos_distance_sq(p2, p1))


def pos_distance_sq(p1, p2):
    """Return the squared euclidean norm of the difference of the two vectors."""
    x = p1.x - p2.x
    y = p1.y - p2.y
    z = p1.z - p2.z
    return x * x + y * y + z * z


def vec_angle(a, b):
    """Get the angle between the two vectors."""
    a_len = a.length()
    b_len = b.length()
    if a_len * b_len < 1E-6:
        return 0.0
    c = a.dot(b) / a_len / b_len
    if c >= 1.0:
        return 0.0
    elif c <= -1.0:
        return math.pi
    else:
        return math.acos(c)


def pos_angle(p1, p2, p3):
    """Get the angle from three points."""
    return vec_angle(p1 - p2, p3 - p2)


def vec_dihedral(a, axis, c):
    """Get the dihedral angle between a and c along the axis."""
    ap = a.perpendicular_vec(axis)
    cp = c.perpendicular_vec(axis)
    angle = vec_angle(ap, cp)
    if ap.cross(cp).dot(axis) > 0:
        angle = -angle
    return angle


def pos_dihedral(p1, p2, p3, p4):
    """Get the dihedral angle from four points."""
    return vec_dihedral(p1 - p2, p2 - p3, p4 - p3)


def rotated_pos(theta, anchor, center, pos):
    """Rotate the position by theta around the center."""
    return rotation(center - anchor, theta, center).transform_vec(pos)


# def ProjectedPos(length, angle, dihedral, p1, p2, p3):
#     """Project."""
#     norm = plane_normal(p1, p2, p3)
#     axis = p3 - p2
#     vec_diff = axis.scaled_vec(-length)
#     vec_diff = RotationAtOrigin(norm, -angle).transform_vec(vec_diff)
#     vec_diff = RotationAtOrigin(axis, dihedral).transform_vec(vec_diff)
    # return p3 + vec_diff


class Matrix3D(np.matrix):
    """Object to manipulate three-dimensional matrices.

    Wrapper around numpy matrix object with shape (4, 4).
    """

    def __new__(cls, values: Iterable[float] = None, order='C') -> Matrix3D:
        """Create a new Matrix3D object.

        :param values: values of the matrix
        :param order: index order of provided values
                      'C': column-first indexing (first items of values will make first row)
                      'F': row-first indexing (first items of values will make first column)
        """
        if values is None:
            self = np.zeros((4, 4), dtype=float)
        else:
            len_ = len(list(values))
            if not len_ == 16:
                raise ValueError(f'Matrix3D cannot be instanciated with {len_} values')
            self = np.array(values, dtype=float).reshape((4, 4), order=order)
        return self.view(cls)

    def transform_vector(self, vector: Vector3D) -> Vector3D:
        """Apply the transformation to the vector.

        :param vector: vector to be transformed
        """
        return Vector3D.from_numpy(self[:3, :3].T * vector.to_numpy_col() + self[3, :3].T)


def rotation_at_origin(axis, theta):
    """Create a matrix for a rotation around the origin.

    :param axis: axis around which the rotation should be performed
    :param theta: angle of the rotation
    """
    v = axis.normal_vec()
    c = math.cos(float(theta))
    s = math.sin(float(theta))
    t = 1.0 - c
    m = Matrix3D([[t * v.x * v.x + c,
                   t * v.x * v.y + v.z * s,
                   t * v.x * v.z - v.y * s,
                   0],
                  [t * v.y * v.x - v.z * s,
                   t * v.y * v.y + c,
                   t * v.y * v.z + v.x * s,
                   0],
                  [t * v.z * v.x + v.y * s,
                   t * v.z * v.y - v.x * s,
                   t * v.z * v.z + c,
                   0],
                  [0, 0, 0, 0]])
    return m


def translation(vector):
    """Create a translation matrix.

    :param vector: vector of translation
    """
    m = Matrix3D()
    m[3, :3] = [vector.x, vector.y, vector.z]
    return m


def rotation(axis, theta, center):
    """Create the matrix to rotate around an axis at center."""
    rot = rotation_at_origin(axis, theta)
    trans = translation(center - rot.transform_vec(center))
    return trans * rot


def superposition3(ref1, ref2, ref3, mov1, mov2, mov3):
    """Superimpose."""
    mov_diff = mov2 - mov1
    ref_diff = ref2 - ref1

    m1 = Matrix3D()
    if math.fabs(vec_angle(mov_diff, ref_diff)) < SMALL:
        m1 = translation(ref1 - mov1)
    else:
        axis = mov_diff.cross(ref_diff)
        torsion = vec_dihedral(ref_diff, axis, mov_diff)
        rot = rotation_at_origin(axis, torsion)
        trans = translation(ref2 - rot.transform_vec(mov2))
        m1 = trans * rot

    mov_diff = ref2 - m1.transform_vec(mov3)
    ref_diff = ref2 - ref3

    m = Matrix3D()
    if math.fabs(vec_angle(mov_diff, ref_diff)) < SMALL:
        m = m1
    else:
        axis = ref2 - ref1
        torsion = vec_dihedral(ref_diff, axis, mov_diff)
        m2 = rotation_at_origin(axis, torsion)
        m3 = translation(ref2 - m2.transform_vec(ref2))
        m = m3 * m2 * m1

    return m


def random_vec():
    """Get a random vector."""
    return Vector3D(random.uniform(-100, 100),  # noqa: S311
                    random.uniform(-100, 100),  # noqa: S311
                    random.uniform(-100, 100))  # noqa: S311


def random_origin_rotation():
    """Get a random rotation around the origin."""
    axis = random_vec()
    angle = random.uniform(-math.pi / 2, math.pi / 2)  # noqa: S311
    return rotation_at_origin(axis, angle)


def random_transform():
    """Get a random transformation matrix."""
    axis = random_vec()
    angle = random.uniform(-math.pi / 2, math.pi / 2)  # noqa: S311
    center = random_vec()
    return rotation(axis, angle, center)



# -*- coding: utf-8 -*-


"""Basak information topological indices."""


import copy

import numpy as np
from rdkit import Chem

from .topology import Topology


class Basak:
    """Basak information content descriptors."""
    
    @staticmethod
    def calculate_basak_ic0(mol: Chem.Mol)-> float:
        """Calculate information content of order 0."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = []
        for i in range(nAtoms):
            at = Hmol.GetAtomWithIdx(i)
            IC.append(at.GetAtomicNum())
        Unique = np.unique(IC)
        NAtomType = len(Unique)
        NTAtomType = np.zeros(NAtomType, dtype=float)
        for i in range(NAtomType):
            NTAtomType[i] = IC.count(Unique[i])
        if nAtoms != 0:
            BasakIC = Topology._calculate_entropy(NTAtomType / nAtoms)
        else:
            BasakIC = 0.0
        return BasakIC

    @staticmethod
    def calculate_basak_sic0(mol: Chem.Mol)-> float:
        """Calculate the structural information content of order 0."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic0(mol)
        if nAtoms <= 1:
            BasakSIC = 0.0
        else:
            BasakSIC = IC / np.log2(nAtoms)
        return BasakSIC

    @staticmethod
    def calculate_basak_cic0(mol: Chem.Mol)-> float:
        """Calculate the complementary information content of order 0."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic0(mol)
        if nAtoms <= 1:
            BasakCIC = 0.0
        else:
            BasakCIC = np.log2(nAtoms) - IC
        return BasakCIC

    @staticmethod
    def _calculate_basak_ic_n(mol: Chem.Mol, NumPath=1)-> float:
        """Calculate the information content of order n."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        TotalPath = Chem.FindAllPathsOfLengthN(Hmol, NumPath, useBonds=0, useHs=1)
        if len(TotalPath) == 0:
            BasakIC = 0.0
        else:
            IC = {}
            for i in range(nAtoms):
                temp = []
                at = Hmol.GetAtomWithIdx(i)
                temp.append(at.GetAtomicNum())
                for index in TotalPath:
                    if i == index[0]:
                        temp.append([Hmol.GetAtomWithIdx(kk).GetAtomicNum() for kk in index[1:]])
                    if i == index[-1]:
                        cds = list(index)
                        cds.reverse()
                        temp.append([Hmol.GetAtomWithIdx(kk).GetAtomicNum() for kk in cds[1:]])
                IC[str(i)] = temp
            cds = []
            for value in IC.values():
                cds.append(value)
            kkk = list(range(len(cds)))
            aaa = copy.deepcopy(kkk)
            res = []
            for i in aaa:
                if i in kkk:
                    jishu = 0
                    kong = []
                    temp1 = cds[i]
                    for j in aaa:
                        if cds[j] == temp1:
                            jishu += 1
                            kong.append(j)
                    for ks in kong:
                        kkk.remove(ks)
                    res.append(jishu)
            BasakIC = Topology._calculate_entropy(np.array(res, dtype=float) / sum(res))
        return BasakIC

    @staticmethod
    def calculate_basak_ic1(mol: Chem.Mol)-> float:
        """Calculate the information content of order 1."""
        return Basak._calculate_basak_ic_n(mol, NumPath=2)

    @staticmethod
    def calculate_basak_ic2(mol: Chem.Mol)-> float:
        """Calculate the information content of order 2."""
        return Basak._calculate_basak_ic_n(mol, NumPath=3)

    @staticmethod
    def calculate_basak_ic3(mol: Chem.Mol)-> float:
        """Calculate the information content of order 3."""
        return Basak._calculate_basak_ic_n(mol, NumPath=4)

    @staticmethod
    def calculate_basak_ic4(mol: Chem.Mol)-> float:
        """Calculate the information content of order 4."""
        return Basak._calculate_basak_ic_n(mol, NumPath=5)

    @staticmethod
    def calculate_basak_ic5(mol: Chem.Mol)-> float:
        """Calculate the information content of order 5."""
        return Basak._calculate_basak_ic_n(mol, NumPath=6)

    @staticmethod
    def calculate_basak_ic6(mol: Chem.Mol)-> float:
        """Calculate the information content of order 6."""
        return Basak._calculate_basak_ic_n(mol, NumPath=7)

    @staticmethod
    def calculate_basak_sic1(mol: Chem.Mol)-> float:
        """Calculate the structural information content of order 1."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic1(mol)
        if nAtoms <= 1:
            BasakSIC = 0.0
        else:
            BasakSIC = IC / np.log2(nAtoms)
        return BasakSIC

    @staticmethod
    def calculate_basak_sic2(mol: Chem.Mol)-> float:
        """Calculate the structural information content of order 2."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic2(mol)
        if nAtoms <= 1:
            BasakSIC = 0.0
        else:
            BasakSIC = IC / np.log2(nAtoms)
        return BasakSIC

    @staticmethod
    def calculate_basak_sic3(mol: Chem.Mol)-> float:
        """Calculate the structural information content of order 3."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic3(mol)
        if nAtoms <= 1:
            BasakSIC = 0.0
        else:
            BasakSIC = IC / np.log2(nAtoms)
        return BasakSIC

    @staticmethod
    def calculate_basak_sic4(mol: Chem.Mol)-> float:
        """Calculate the structural information content of order 4."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic4(mol)
        if nAtoms <= 1:
            BasakSIC = 0.0
        else:
            BasakSIC = IC / np.log2(nAtoms)
        return BasakSIC

    @staticmethod
    def calculate_basak_sic5(mol: Chem.Mol)-> float:
        """Calculate the structural information content of order 5."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic5(mol)
        if nAtoms <= 1:
            BasakSIC = 0.0
        else:
            BasakSIC = IC / np.log2(nAtoms)
        return BasakSIC

    @staticmethod
    def calculate_basak_sic6(mol: Chem.Mol)-> float:
        """Calculate the structural information content of order 6."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic6(mol)
        if nAtoms <= 1:
            BasakSIC = 0.0
        else:
            BasakSIC = IC / np.log2(nAtoms)
        return BasakSIC

    @staticmethod
    def calculate_basak_cic1(mol: Chem.Mol)-> float:
        """Calculate the complementary information content of order 1."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic1(mol)
        if nAtoms <= 1:
            BasakCIC = 0.0
        else:
            BasakCIC = np.log2(nAtoms) - IC
        return BasakCIC

    @staticmethod
    def calculate_basak_cic2(mol: Chem.Mol)-> float:
        """Calculate the complementary information content of order 2."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic2(mol)
        if nAtoms <= 1:
            BasakCIC = 0.0
        else:
            BasakCIC = np.log2(nAtoms) - IC
        return BasakCIC

    @staticmethod
    def calculate_basak_cic3(mol: Chem.Mol)-> float:
        """Calculate the complementary information content of order 3."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic3(mol)
        if nAtoms <= 1:
            BasakCIC = 0.0
        else:
            BasakCIC = np.log2(nAtoms) - IC
        return BasakCIC

    @staticmethod
    def calculate_basak_cic4(mol: Chem.Mol)-> float:
        """Calculate the complementary information content of order 4."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic4(mol)
        if nAtoms <= 1:
            BasakCIC = 0.0
        else:
            BasakCIC = np.log2(nAtoms) - IC
        return BasakCIC

    @staticmethod
    def calculate_basak_cic5(mol: Chem.Mol)-> float:
        """Calculate the complementary information content of order 5."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic5(mol)
        if nAtoms <= 1:
            BasakCIC = 0.0
        else:
            BasakCIC = np.log2(nAtoms) - IC
        return BasakCIC

    @staticmethod
    def calculate_basak_cic6(mol: Chem.Mol)-> float:
        """Calculate the complementary information content of order 6."""
        Hmol = Chem.AddHs(mol)
        nAtoms = Hmol.GetNumAtoms()
        IC = Basak.calculate_basak_ic6(mol)
        if nAtoms <= 1:
            BasakCIC = 0.0
        else:
            BasakCIC = np.log2(nAtoms) - IC
        return BasakCIC
    
    
    _basak = {'IC0': calculate_basak_ic0,
              'IC1': calculate_basak_ic1,
              'IC2': calculate_basak_ic2,
              'IC3': calculate_basak_ic3,
              'IC4': calculate_basak_ic4,
              'IC5': calculate_basak_ic5,
              'IC6': calculate_basak_ic6,
              'SIC0': calculate_basak_sic0,
              'SIC1': calculate_basak_sic1,
              'SIC2': calculate_basak_sic2,
              'SIC3': calculate_basak_sic3,
              'SIC4': calculate_basak_sic4,
              'SIC5': calculate_basak_sic5,
              'SIC6': calculate_basak_sic6,
              'CIC0': calculate_basak_cic0,
              'CIC1': calculate_basak_cic1,
              'CIC2': calculate_basak_cic2,
              'CIC3': calculate_basak_cic3,
              'CIC4': calculate_basak_cic4,
              'CIC5': calculate_basak_cic5,
              'CIC6': calculate_basak_cic6,
              }

    @staticmethod
    def get_all(mol: Chem.Mol) -> dict:
        """Calculate all (21) Basak descriptors."""
        result = {}
        for DesLabel in Basak._basak.keys():
            result[DesLabel] = Basak._basak[DesLabel](mol)
        return result



# -*- coding: utf-8 -*-


"""Charge descriptors based on Gasteiger/Marseli partial charges."""

import numpy as np
from rdkit import Chem
from rdkit.Chem import rdPartialCharges as GMCharge


iter_step = 12


class Charge:
    """Charge descriptors."""
    
    @staticmethod
    def _calculate_element_max_pcharge(mol: Chem.Mol, AtomicNum: int = 6) -> float:
        """Get the most positive charge of atom with specified atomic number."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            if atom.GetAtomicNum() == AtomicNum:
                res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            return max(res)

    @staticmethod
    def _calculate_element_max_ncharge(mol: Chem.Mol, AtomicNum: int = 6) -> float:
        """Get the most negative charge of atom with specified atomic number."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            if atom.GetAtomicNum() == AtomicNum:
                res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            return min(res)

    @staticmethod
    def calculate_h_max_pcharge(mol: Chem.Mol) -> float:
        """Get most positive charge of all hydrogen atoms."""
        return Charge._calculate_element_max_pcharge(mol, AtomicNum=1)

    @staticmethod
    def calculate_c_max_pcharge(mol: Chem.Mol) -> float:
        """Get most positive charge of all carbon atoms."""
        return Charge._calculate_element_max_pcharge(mol, AtomicNum=6)

    @staticmethod
    def calculate_n_max_pcharge(mol: Chem.Mol) -> float:
        """Get most positive charge of all nitrogen atoms."""
        return Charge._calculate_element_max_pcharge(mol, AtomicNum=7)

    @staticmethod
    def calculate_o_max_pcharge(mol: Chem.Mol) -> float:
        """Get most positive charge of all oxygen atoms."""
        return Charge._calculate_element_max_pcharge(mol, AtomicNum=8)

    @staticmethod
    def calculate_h_max_ncharge(mol) -> float:
        """Get most negative charge of all hydrogen atoms."""
        return Charge._calculate_element_max_ncharge(mol, AtomicNum=1)

    @staticmethod
    def calculate_c_max_ncharge(mol: Chem.Mol) -> float:
        """Get most negative charge of all carbon atoms."""
        return Charge._calculate_element_max_ncharge(mol, AtomicNum=6)

    @staticmethod
    def calculate_n_max_ncharge(mol: Chem.Mol) -> float:
        """Get most negative charge of all nitrogen atoms."""
        return Charge._calculate_element_max_ncharge(mol, AtomicNum=7)

    @staticmethod
    def calculate_o_max_ncharge(mol: Chem.Mol) -> float:
        """Get most negative charge of all oxygen atoms."""
        return Charge._calculate_element_max_ncharge(mol, AtomicNum=8)

    @staticmethod
    def calculate_all_max_pcharge(mol: Chem.Mol) -> float:
        """Get most positive charge of all atoms."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            return max(res)

    @staticmethod
    def calculate_all_max_ncharge(mol: Chem.Mol) -> float:
        """Get most negative charge of all atoms."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            return min(res)

    @staticmethod
    def _calculate_element_sum_square_charge(mol: Chem.Mol, AtomicNum: int = 6) -> float:
        """Get the sum of square charges of atoms with specified atomic number."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            if atom.GetAtomicNum() == AtomicNum:
                res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            return sum(np.square(res))

    @staticmethod
    def calculate_h_sum_square_charge(mol: Chem.Mol) -> float:
        """Get the sum of square charges of hydrogen atoms."""
        return Charge._calculate_element_sum_square_charge(mol, AtomicNum=1)

    @staticmethod
    def calculate_c_sum_square_charge(mol: Chem.Mol) -> float:
        """Get the sum of square charges of carbon atoms."""
        return Charge._calculate_element_sum_square_charge(mol, AtomicNum=6)

    @staticmethod
    def calculate_n_sum_square_charge(mol: Chem.Mol) -> float:
        """Get the sum of square charges of nitrogen atoms."""
        return Charge._calculate_element_sum_square_charge(mol, AtomicNum=7)

    @staticmethod
    def calculate_o_sum_square_charge(mol: Chem.Mol) -> float:
        """Get the sum of square charges of oxygen atoms."""
        return Charge._calculate_element_sum_square_charge(mol, AtomicNum=8)

    @staticmethod
    def calculate_all_sum_square_charge(mol: Chem.Mol)-> float:
        """Get the sum of square charges of all atoms."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            return sum(np.square(res))

    @staticmethod
    def calculate_total_pcharge(mol: Chem.Mol) -> float:
        """Get the total positive charge."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            cc = np.array(res, 'd')
            return sum(cc[cc > 0])

    @staticmethod
    def calculate_mean_pcharge(mol: Chem.Mol) -> float:
        """Get the average positive charge."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            cc = np.array(res, 'd')
            cc = cc[cc > 0]
            return np.mean(cc) if len(cc) else np.nan

    @staticmethod
    def calculate_total_ncharge(mol: Chem.Mol) -> float:
        """Ge the total negative charge."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            cc = np.array(res, 'd')
            cc = cc[cc < 0]
            return sum(cc) if len(cc) else np.nan

    @staticmethod
    def calculate_mean_ncharge(mol: Chem.Mol) -> float:
        """Get the average negative charge."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            cc = np.array(res, 'd')
            cc = cc[cc < 0]
            return np.mean(cc) if len(cc) else np.nan

    @staticmethod
    def calculate_total_absolute_charge(mol: Chem.Mol) -> float:
        """Get the total absolute charge."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            cc = np.array(res, 'd')
            return sum(np.absolute(cc))

    @staticmethod
    def calculate_mean_absolute_charge(mol: Chem.Mol) -> float:
        """Get the average absolute charge."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            cc = np.array(res, 'd')
            return np.mean(np.absolute(cc))

    @staticmethod
    def calculate_relative_pcharge(mol: Chem.Mol) -> float:
        """Get the ratio between the most positive partial charge and the total positive charge."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            cc = np.array(res, 'd')
            if sum(cc[cc > 0]) == 0:
                return 0
            else:
                return max(res) / sum(cc[cc > 0])

    @staticmethod
    def calculate_relative_ncharge(mol: Chem.Mol) -> float:
        """Get the ratio between the most negative partial charge and the total negative charge."""
        Hmol = Chem.AddHs(mol)
        GMCharge.ComputeGasteigerCharges(Hmol, iter_step)
        res = []
        for atom in Hmol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        if res == []:
            return 0
        else:
            cc = np.array(res, 'd')
            if sum(cc[cc < 0]) == 0:
                return 0
            else:
                return min(res) / sum(cc[cc < 0])

    @staticmethod
    def calculate_local_dipole_index(mol: Chem.Mol) -> float:
        """calculate_ the local dipole index (D)."""
        GMCharge.ComputeGasteigerCharges(mol, iter_step)
        res = []
        for atom in mol.GetAtoms():
            res.append(float(atom.GetProp('_GasteigerCharge')))
        cc = [np.absolute(res[x.GetBeginAtom().GetIdx()] - res[x.GetEndAtom().GetIdx()]) for x in mol.GetBonds()]
        B = len(mol.GetBonds())
        return 0 if len(cc) == 0.0 else sum(cc) / B

    @staticmethod
    def calculate_submol_polarity_param(mol: Chem.Mol) -> float:
        """calculate_ the submolecular polarity parameter (SPP)."""
        return Charge.calculate_all_max_pcharge(mol) - Charge.calculate_all_max_ncharge(mol)
    
    
    _charge = {'QHmax': calculate_h_max_pcharge,
               'QCmax': calculate_c_max_pcharge,
               'QNmax': calculate_n_max_pcharge,
               'QOmax': calculate_o_max_pcharge,
               'QHmin': calculate_h_max_ncharge,
               'QCmin': calculate_c_max_ncharge,
               'QNmin': calculate_n_max_ncharge,
               'QOmin': calculate_o_max_ncharge,
               'Qmax': calculate_all_max_pcharge,
               'Qmin': calculate_all_max_ncharge,
               'QHss': calculate_h_sum_square_charge,
               'QCss': calculate_c_sum_square_charge,
               'QNss': calculate_n_sum_square_charge,
               'QOss': calculate_o_sum_square_charge,
               'Qass': calculate_all_sum_square_charge,
               'Mpc': calculate_mean_pcharge,
               'Tpc': calculate_total_pcharge,
               'Mnc': calculate_mean_ncharge,
               'Tnc': calculate_total_ncharge,
               'Mac': calculate_mean_absolute_charge,
               'Tac': calculate_total_absolute_charge,
               'Rnc': calculate_relative_ncharge,
               'Rpc': calculate_relative_pcharge,
               'SPP': calculate_submol_polarity_param,
               'LDI': calculate_local_dipole_index,
               }

    @staticmethod
    def get_all(mol: Chem.Mol) -> dict:
        """Get all (25) constitutional descriptors."""
        result = {}
        for DesLabel in Charge._charge.keys():
            result[DesLabel] = Charge._charge[DesLabel](mol)
        return result



# -*- coding: utf-8 -*-


"""Quantum chemistry descriptors."""

from typing import List, Tuple

from rdkit import Chem
import numpy as np


class QuantumChemistry:
    """Quantum chemistry descriptors."""

    @staticmethod
    def _get_max(x: float) -> float:
        """Get the maximum of x.

        If x is empty return 0.0.
        """
        if x == []:
            return 0.0
        else:
            return max(x)

    @staticmethod
    def _get_min(x: float) -> float:
        """Get the minimum of x.

        Ff x is empty return 0.0.
        """
        if x == []:
            return 0.0
        else:
            return min(x)

    @staticmethod
    def read_file(filename: str) -> dict:
        """Read basic quantum chemistry descriptors from an .arc file."""
        inputdict = {'Hf': np.NaN, 'ET': np.NaN, 'mu': np.NaN, 'EHomo': np.NaN, 'ELumo': np.NaN,
                     'ESomo': np.NaN, 'EaSomo': np.NaN, 'EaLumo': np.NaN, 'EbSomo': np.NaN, 'EbLumo': np.NaN,
                     'mMw': np.NaN, 'CoArea': np.NaN, 'CoVolume': np.NaN}
        f = open(filename, 'r')
        for line in f.readlines():
            value = line[10:34].strip()
            if value == "HEAT OF FORMATION":
                # 1ev = 96.4853 kj/mol
                inputdict['Hf'] = float(line.strip().upper().split('=')[-1].upper().strip('KJ/MOL')) / 96.4853
            if value == "TOTAL ENERGY":
                inputdict['ET'] = float(line.strip().upper().split('=')[1].upper().strip('EV'))
            if value == "DIPOLE":
                inputdict['mu'] = float(line.strip().upper().split('=')[1].split('DEBYE')[0])
            if value.startswith("HOMO LUMO ENERGIES"):
                inputdict['EHomo'] = float(line.split('=')[1].strip().split()[0])
                inputdict['ELumo'] = float(line.split('=')[1].strip().split()[1])
            elif value.startswith("HOMO (SOMO) LUMO"):
                data = line.split('=')[1].strip().translate(str.maketrans('', '', '()')).split()
                data = list(filter(None, data))
                inputdict['EHomo'] = float(data[0])
                inputdict['ESomo'] = float(data[1])
                inputdict['ELumo'] = float(data[2])
            elif value.startswith("ALPHA SOMO LUMO"):
                data = line.split('=')[1].strip().translate(str.maketrans('', '', '()')).split()
                data = list(filter(None, data))
                inputdict['EaSomo'] = float(data[0])
                inputdict['EaLumo'] = float(data[1])
            elif value.startswith("BETA  SOMO LUMO"):
                data = line.split('=')[1].strip().translate(str.maketrans('', '', '()')).split()
                data = list(filter(None, data))
                inputdict['EbSomo'] = float(data[0])
                inputdict['EbLumo'] = float(data[1])
            if line[10:26] == "MOLECULAR WEIGHT":
                inputdict['mMw'] = float(line[-12:-1])
            elif value == "COSMO AREA":
                inputdict['CoArea'] = float(line.split('=')[1].strip().split()[0])
            elif value == "COSMO VOLUME":
                inputdict['CoVolume'] = float(line.split('=')[1].strip().split()[0])
        f.close()
        return inputdict

    @staticmethod
    def _read_charge(arc_file: str) -> List[Tuple[str, float]]:
        """Read the charge of each atom in .arc file.

        :param arc_file: Path to MOPAC .arc file
        """
        charge = []
        with open(arc_file, 'r') as f:
            templine = f.readlines()

        for line in range(len(templine)):
            if templine[line][-7:-1] == "CHARGE":
                k = line
                break

        for i in templine[k + 4: len(templine) - 1]:
            temp = i.split()
            charge.append((temp[0].strip(), temp[-1].strip()))
        return charge

    @staticmethod
    def get_charge_descriptors(arc_file: str) -> dict:
        """Calculate charge descriptors.

        :param arc_file: Path to MOPAC .arc file
        """
        res = {}
        Htemp = []
        Ctemp = []
        Ntemp = []
        Otemp = []
        temp = []
        Charge = QuantumChemistry._read_charge(arc_file)
        for i in Charge:
            temp.append(float(i[1]))
            if i[0] == 'H':
                Htemp.append(float(i[1]))
            if i[0] == 'C':
                Ctemp.append(float(i[1]))
            if i[0] == 'N':
                Ntemp.append(float(i[1]))
            if i[0] == 'O':
                Otemp.append(float(i[1]))
        res['qQHmax'] = QuantumChemistry._get_max(Htemp)
        res['qQCmax'] = QuantumChemistry._get_max(Ctemp)
        res['qQNmax'] = QuantumChemistry._get_max(Ntemp)
        res['qQOmax'] = QuantumChemistry._get_max(Otemp)
        res['qQHmin'] = QuantumChemistry._get_min(Htemp)
        res['qQCmin'] = QuantumChemistry._get_min(Ctemp)
        res['qQNmin'] = QuantumChemistry._get_min(Ntemp)
        res['qQOmin'] = QuantumChemistry._get_min(Otemp)
        res['qQmax'] = max(temp)
        res['qQmin'] = min(temp)
        res['qQHss'] = sum(i * i for i in Htemp)
        res['qQCss'] = sum(i * i for i in Ctemp)
        res['qQNss'] = sum(i * i for i in Ntemp)
        res['qQOss'] = sum(i * i for i in Otemp)
        res['qQass'] = sum(i * i for i in temp)
        res['qMpc'] = np.mean([i for i in temp if i > 0])
        res['qTpc'] = sum(i for i in temp if i > 0)
        res['qMnc'] = np.mean([i for i in temp if i < 0])
        res['qTnc'] = sum(i for i in temp if i < 0)
        res['qMac'] = np.mean([np.abs(i) for i in temp])
        res['qTac'] = sum(np.abs(i) for i in temp)
        res['qRpc'] = QuantumChemistry._get_max(temp) / res['qTpc']
        res['qRnc'] = QuantumChemistry._get_min(temp) / res['qTnc']
        return res

    @staticmethod
    def calculate_basic_quantum_chemistry(inputdict: dict) -> dict:
        """Calculate between 38 and 40 quantum chemical descriptors.

        Derived from Lumo, Homo, dipole moment, enthalpy and the total energy.
        """
        if inputdict.get('EHomo') is not None:
            EHomo = inputdict['EHomo']
        else:
            EHomo = inputdict['EaSomo']
        if inputdict.get('ELumo') is not None:
            ELumo = inputdict['ELumo']
        else:
            ELumo = inputdict['EaLumo']
        dict_ = {}
        dict_.update(inputdict)
        dict_['GAP'] = ELumo - EHomo
        dict_['S'] = 2. / (ELumo - EHomo)
        dict_['eta'] = (ELumo - EHomo) / 2.0
        dict_['fHL'] = EHomo / ELumo
        dict_['IP'] = -EHomo
        dict_['EA'] = -ELumo
        dict_['xmu'] = (-ELumo - EHomo) / 2.0
        return dict_

    @staticmethod
    def get_all(mol: Chem.Mol, arc_file: str) -> dict:
        """Get all quantum chemistry descriptors.

        :param arc_file: Path to MOPAC .arc file
        """
        inputdict = QuantumChemistry.read_file(arc_file)
        res = QuantumChemistry.calculate_basic_quantum_chemistry(inputdict)
        res.update(QuantumChemistry.get_charge_descriptors(arc_file))
        return res



# -*- coding: utf-8 -*-


"""Geometrical molecular descriptors.

Optimizes molecular structure using MOPAC.
"""

import math
from typing import List, Tuple

import numpy as np
import numpy.lib.scimath
from rdkit import Chem
from rdkit.Chem.Descriptors import ExactMolWt

from .topology import periodicTable
from .geo_opt import read_coordinates
from .utils import get_atom_distance, get_geometrical_distance_matrix


class Geometric:
    """Geometrical molecular descriptors."""

    @staticmethod
    def _get_center_of_mass(mass_coordinates: List[Tuple[float, Tuple[float, float, float]]]) -> Tuple[float, float, float]:
        """Get the center of mass.

        :param mass_coordinates: list of atomic masses and coordinates
                                in the format [(atommass, (x,y,z)), ...]
        """
        res1 = 0.0
        res2 = 0.0
        res3 = 0.0
        temp = []
        for i in mass_coordinates:
            res1 += i[0] * i[1][0]
            res2 += i[0] * i[1][1]
            res3 += i[0] * i[1][2]
            temp.append(i[0])
        result = (res1 / sum(temp), res2 / sum(temp), res3 / sum(temp))
        return result

    @staticmethod
    def _get_geometrical_center(charge_coordinates: List[List[float]]) -> Tuple[float, float, float]:
        """Get the geometrical center.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        res1 = []
        res2 = []
        res3 = []
        for i in charge_coordinates:
            res1.append(float(i[1]))
            res2.append(float(i[2]))
            res3.append(float(i[3]))
        result = (np.mean(res1), np.mean(res2), np.mean(res3))
        return result

    @staticmethod
    def calculate_3D_wiener_with_h(charge_coordinates: List[List[float]]) -> float:
        """Calculate 3D Wiener index from geometrical distance matrix of a MOPAC optimized molecule (including Hs).

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        temp = []
        for i in charge_coordinates:
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        distance_matrix = get_geometrical_distance_matrix(temp)
        return np.sum(distance_matrix) / 2.0

    @staticmethod
    def calculate_3D_wiener_without_h(charge_coordinates: List[List[float]]) -> float:
        """Calculate 3D Wiener index from geometrical distance matrix of a MOPAC optimized molecule (not including Hs).

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        temp = []
        for i in charge_coordinates:
            if i[0] != 'H':
                temp.append([float(i[1]), float(i[2]), float(i[3])])
        distance_matrix = get_geometrical_distance_matrix(temp)
        return np.sum(distance_matrix) / 2.0

    @staticmethod
    def calculate_petitjean_3D_index(charge_coordinates: List[List[float]]) -> float:
        """Calculate Petitjean Index from molecular gemetrical distance matrix.

        The 3D Petitjean shape index (PJI3) is calculated
        dividing the difference between geometric diameter and
        radius by the geometric radius [P.A. Bath, A.R. Poirrette,
        P. Willett, F.H. Allen, J.Chem.Inf.Comput.Sci. 1995, 35, 714-716].
        The geometric radius of a molecule is defined as the minimum
        geometric eccentricity and the diameter is defined as the
        maximum geometric eccentricity in the molecule, the atom
        geometric eccentricity being the longest geometric distance
        from the considered atom to any other atom in the molecule.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        temp = []
        for i in charge_coordinates:
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        distance_matrix = get_geometrical_distance_matrix(temp)
        temp1 = np.amax(distance_matrix, axis=0)
        return max(temp1) / min(temp1) - 1.0

    @staticmethod
    def calculate_geometrical_diameter(charge_coordinates: List[List[float]]) -> float:
        """Calculate the geometrical diameter.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        temp = []
        for i in charge_coordinates:
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        distance_matrix = get_geometrical_distance_matrix(temp)
        temp1 = np.amax(distance_matrix, axis=0)
        return max(temp1)

    @staticmethod
    def calculate_topo_electronic(charge_coordinates: List[List[float]]) -> float:
        """Calculate Topographic electronic descriptors.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        temp, charges = [], []
        for i in charge_coordinates:
            temp.append([float(i[1]), float(i[2]), float(i[3])])
            charges.append(float(i[4]))
        nAT = len(charge_coordinates)
        result = 0.0
        for i in range(nAT - 1):
            for j in range(i + 1, nAT):
                dis = get_atom_distance(temp[i], temp[j])
                result += np.absolute(charges[i] * charges[j]) / numpy.lib.scimath.power(dis, p=2)
        return result

    @staticmethod
    def calculate_gravitational_3D1(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> float:
        """Calculate Gravitational 3D index from all atoms.

        :param mol: molecule
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        mol = Chem.AddHs(mol)
        temp = []
        atoms = list(mol.GetAtoms())
        for i, j in enumerate(charge_coordinates):
            temp.append([periodicTable.GetAtomicWeight(atoms[i].GetAtomicNum()),
                         [float(j[1]), float(j[2]), float(j[3])]])
        nAT = len(temp)
        result = 0.0
        for i in range(nAT - 1):
            for j in range(i + 1, nAT):
                dis = get_atom_distance(temp[i][1], temp[j][1])
                result += temp[i][0] * temp[j][0] / numpy.lib.scimath.power(dis, p=2)
        return float(result) / 100

    @staticmethod
    def calculate_gravitational_3D2(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> float:
        """Calculate Gravitational 3D index from bonded atoms.

        Katritzky, A.R. et al., J.Phys.Chem. 1996, 100, 10400-10407]
        :param mol: molecule
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        raise NotImplementedError('Needs modification: calculated only on bonded atoms.')
        mol = Chem.AddHs(mol)
        temp = []
        atoms = list(mol.GetAtoms())
        for i, j in enumerate(charge_coordinates):
            temp.append([periodicTable.GetAtomicWeight(atoms[i].GetAtomicNum()),
                         [float(j[1]), float(j[2]), float(j[3])]])
        nAT = len(temp)
        result = 0.0
        for i in range(nAT - 1):
            for j in range(i + 1, nAT):
                dis = get_atom_distance(temp[i][1], temp[j][1])
                result += temp[i][0] * temp[j][0] / numpy.lib.scimath.power(dis, p=2)
        return float(result) / 100

    @staticmethod
    def calculate_radius_of_gyration(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> float:
        """Calculate Radius of gyration.

        :param mol: molecule
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        mol = Chem.AddHs(mol)
        temp = []
        atoms = list(mol.GetAtoms())
        for i, j in enumerate(charge_coordinates):
            temp.append([periodicTable.GetAtomicWeight(atoms[i].GetAtomicNum()),
                         [float(j[1]), float(j[2]), float(j[3])]])
        nAT = len(temp)
        com = Geometric._get_center_of_mass(temp)
        result = 0.0
        for i in range(nAT):
            dis = get_atom_distance(temp[i][1], com)
            result += temp[i][0] * numpy.lib.scimath.power(dis, p=2)
        return np.sqrt(float(result / ExactMolWt(mol)))

    @staticmethod
    def get_inertia_matrix(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> np.array:
        """Get Inertia matrix based on atomic mass and optimized coordinates.

        :param mol: molecule
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        mol = Chem.AddHs(mol)
        temp = []
        atoms = list(mol.GetAtoms())
        for i, j in enumerate(charge_coordinates):
            temp.append([periodicTable.GetAtomicWeight(atoms[i].GetAtomicNum()),
                         [float(j[1]), float(j[2]), float(j[3])]])
        nAT = len(temp)
        inertia_matrix = np.zeros((3, 3))
        res11 = 0.0
        res22 = 0.0
        res33 = 0.0
        res12 = 0.0
        res13 = 0.0
        res23 = 0.0
        for i in range(nAT):
            res11 += temp[i][0] * (math.pow(temp[i][1][1], 2) + math.pow(temp[i][1][2], 2))
            res22 += temp[i][0] * (math.pow(temp[i][1][0], 2) + math.pow(temp[i][1][2], 2))
            res33 += temp[i][0] * (math.pow(temp[i][1][0], 2) + math.pow(temp[i][1][1], 2))
            res12 += temp[i][0] * (temp[i][1][0] * temp[i][1][1])
            res13 += temp[i][0] * (temp[i][1][0] * temp[i][1][2])
            res23 += temp[i][0] * (temp[i][1][1] * temp[i][1][2])
        inertia_matrix[0, 0] = res11
        inertia_matrix[1, 1] = res22
        inertia_matrix[2, 2] = res33
        inertia_matrix[0, 1] = res12
        inertia_matrix[0, 2] = res13
        inertia_matrix[1, 2] = res23
        inertia_matrix[1, 0] = res12
        inertia_matrix[2, 0] = res13
        inertia_matrix[2, 1] = res23
        return inertia_matrix

    @staticmethod
    def calculate_principal_moment_of_inertia(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> dict:
        """Calculate X, Y and Z-principal geometric moments.

        derived from ADAPT developed by Jurs.
        :param mol: molecule
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        inertia_matrix = Geometric.get_inertia_matrix(mol, charge_coordinates)
        ma = np.mean(inertia_matrix, axis=1)
        ms = np.std(inertia_matrix, axis=1, ddof=1)
        bb = np.ones((3, 1))
        inertia_matrix = (inertia_matrix - bb * ma.T) / (bb * ms.T)
        u, s, v = np.linalg.svd(inertia_matrix)
        res = {}
        res['IA'] = s[2]
        res['IB'] = s[1]
        res['IC'] = s[0]
        return res

    @staticmethod
    def calculate_ratio_pmi(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> dict:
        """Calculate the ratio of principal moment of inertia.

        derived from ADAPT developed by Jurs.
        :param mol: molecule
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        temp = Geometric.calculate_principal_moment_of_inertia(mol, charge_coordinates)
        res = {}
        res['IA/B'] = temp['IA'] / temp['IB']
        res['IA/C'] = temp['IA'] / temp['IC']
        res['IB/C'] = temp['IB'] / temp['IC']
        return res

    @staticmethod
    def calculate_harary_3D(charge_coordinates: List[List[float]]) -> float:
        """Calculate 3D-Harary index as the sum of all the reciprocal geometric distances.

        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        temp = []
        for i in charge_coordinates:
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        distance_matrix = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        res = 0.0
        for i in range(nAT - 1):
            for j in range(i + 1, nAT):
                if distance_matrix[i, j] == 0:
                    cds = 0.0
                else:
                    cds = 1. / distance_matrix[i, j]
                res = res + cds
        return res

    @staticmethod
    def calculate_average_geometrical_distance_degree(charge_coordinates: List[List[float]]) -> float:
        """Calculate the average geometric distance degree (AGDD).

        This is the ratio between the sum of all geometric distance degrees and the atoms.
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        temp = []
        for i in charge_coordinates:
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        distance_matrix = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        res = sum(sum(distance_matrix)) / nAT
        return res

    @staticmethod
    def calculate_abs_eigenvalue_sum_on_geometric_matrix(charge_coordinates: List[List[float]]) -> float:
        """Calculate the absolute eigenvalue sum on geometry matrix (SEig).

        This is the sum of the absolute eigenvalues of the geometry matrix.
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        temp = []
        for i in charge_coordinates:
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        distance_matrix = get_geometrical_distance_matrix(temp)
        u, s, vt = np.linalg.svd(distance_matrix)
        return sum(abs(s))

    @staticmethod
    def calculate_span_r(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> float:
        """Calculate the span R.

        This is defined as the radius of the smallest sphere,
        centred on the centre of mass, completely enclosing all atoms of a molecule.

        Arteca G.A. et al., Molecular Shape Descriptors in Reviews in
        Computational Chemistry - Vol. 9, K.B. Lipkowitz, D. Boyd (Eds.),
        VCH Publishers, New York (NY), pp. 191-253, 1991.
        :param mol: molecule
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        mol = Chem.AddHs(mol)
        temp = []
        atoms = list(mol.GetAtoms())
        for i, j in enumerate(charge_coordinates):
            temp.append([periodicTable.GetAtomicWeight(atoms[i].GetAtomicNum()),
                         [float(j[1]), float(j[2]), float(j[3])]])
        masscenter = Geometric._get_center_of_mass(temp)
        res = []
        for i in temp:
            res.append(get_atom_distance(i[1], masscenter))
        return float(max(res))

    @staticmethod
    def calculate_average_span_r(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> float:
        """Calculate the average span R (SPAM).

        This is the root square of the ratio of SPAN over the number of atoms.
        :param mol: molecule
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        mol = Chem.AddHs(mol)
        temp = []
        atoms = list(mol.GetAtoms())
        for i, j in enumerate(charge_coordinates):
            temp.append([periodicTable.GetAtomicWeight(atoms[i].GetAtomicNum()),
                         [float(j[1]), float(j[2]), float(j[3])]])
        nAT = len(temp)
        masscenter = Geometric._get_center_of_mass(temp)
        res = []
        for i in temp:
            res.append(get_atom_distance(i[1], masscenter))
        return math.pow(float(max(res)) / nAT, 0.5)

    @staticmethod
    def calculate_molecular_eccentricity(mol: Chem.Mol, charge_coordinates: List[List[float]]) -> float:
        """Calculate molecular eccentricity.

        G.A. Arteca, Molecular Shape Descriptors in Reviews
        in Computational Chemistry - Vol. 9, K.B. Lipkowitz, D. Boyd (Eds.),
        VCH Publishers, New York (NY), pp. 191-253, 1991.
        :param mol: molecule
        :param charge_coordinates: Atomic coordinates and charges as read by chemopy.geo_opt.read_coordinates
        """
        inertia_matrix = Geometric.get_inertia_matrix(mol, charge_coordinates)
        u, s, v = np.linalg.svd(inertia_matrix)
        res1 = s[0]
        res3 = s[2]
        res = math.pow(res1 * res1 - res3 * res3, 1. / 2) / res1
        return res

    @staticmethod
    def get_all(mol: Chem.Mol, arc_file: str) -> dict:
        """Get all (20) geometrical descriptors.

        :param mol: the molecule
        :param arc_file: Path to MOPAC .arc file
        """
        res = {}
        charge_coordinates = read_coordinates(arc_file)
        res['W3DH'] = Geometric.calculate_3D_wiener_with_h(charge_coordinates)
        res['W3D'] = Geometric.calculate_3D_wiener_without_h(charge_coordinates)
        res['Petitj3D'] = Geometric.calculate_petitjean_3D_index(charge_coordinates)
        res['GeDi'] = Geometric.calculate_geometrical_diameter(charge_coordinates)
        res['TE1'] = Geometric.calculate_topo_electronic(charge_coordinates)
        res['grav1'] = Geometric.calculate_gravitational_3D1(mol, charge_coordinates) if mol is not None else np.nan
        # res['grav2'] = Geometric.CalculateGravitational3D2(mol, charge_coordinates)
        res['rygr'] = Geometric.calculate_radius_of_gyration(mol, charge_coordinates) if mol is not None else np.nan
        res['Harary3D'] = Geometric.calculate_harary_3D(charge_coordinates)
        res['AGDD'] = Geometric.calculate_average_geometrical_distance_degree(charge_coordinates)
        res['SEig'] = Geometric.calculate_abs_eigenvalue_sum_on_geometric_matrix(charge_coordinates)
        res['SPAN'] = Geometric.calculate_span_r(mol, charge_coordinates) if mol is not None else np.nan
        res['ASPAN'] = Geometric.calculate_average_span_r(mol, charge_coordinates) if mol is not None else np.nan
        res['MEcc'] = Geometric.calculate_molecular_eccentricity(mol, charge_coordinates) if mol is not None else np.nan
        res.update(Geometric.calculate_principal_moment_of_inertia(mol, charge_coordinates)) if mol is not None else np.nan
        res.update(Geometric.calculate_ratio_pmi(mol, charge_coordinates)) if mol is not None else np.nan
        return res



# -*- coding: utf-8 -*-

"""Tests for details about molecular descriptors/fingerprints."""

import unittest

from rdkit import Chem
from rdkit.Chem import AllChem
from chemopy import ChemoPy

from .constants import MOL_MOLECULE


class TestDetailsOfDescriptorsAndFingerprints(unittest.TestCase):
    """Tests molecular descriptors."""

    def setUp(self) -> None:
        """Create the molecular descriptor calculator."""
        self.cmp = ChemoPy(ignore_3D=False)
        mol = Chem.Mol(MOL_MOLECULE)
        _ = AllChem.EmbedMolecule(mol)
        self.desc_names = self.cmp.calculate([mol], show_banner=False).columns.tolist()

    def test_each_descriptor_details(self):
        """Test the details for each descriptor available."""
        for desc_name in self.desc_names:
            details = self.cmp.get_details(desc_name)
            self.assertIsNotNone(details)
            self.assertFalse(details.empty)
            self.assertEqual(details.shape[0], 1)
            self.assertEqual(details.Name.item(), desc_name)

    def test_all_descriptor_details(self):
        """Test the details of all descriptors/fingerprints available."""
        details = self.cmp.get_details()
        self.assertIsNotNone(details)
        self.assertFalse(details.empty)
        # Includes the 12 additional fingerprints
        self.assertEqual(details.shape[0], 1196)



# -*- coding: utf-8 -*-

"""Tests for molecular descriptors."""

import unittest

from chemopy import ChemoPy, Fingerprint, Fingerprint3D

from tests.constants import *


class TestDescriptorsAndFingerprints(unittest.TestCase):
    """Tests molecular descriptors."""

    def setUp(self) -> None:
        """Create the molecular descriptor calculator."""
        self.cmp = ChemoPy(ignore_3D=True)
        self.cmp3d = ChemoPy(ignore_3D=False)
        self.cmpfp = ChemoPy(ignore_3D=True, include_fps=True)
        self.cmp3dfp = ChemoPy(ignore_3D=False, include_fps=True)

    def test_2D_descriptors(self):
        """Test the dimensions of the output descriptor dataframe."""
        values = self.cmp.calculate(MOLECULES_2D, show_banner=False)
        self.assertEqual(values.shape, (len(MOLECULES_2D), 632))
        self.assertEqual(len(values.columns.unique().tolist()), 632)

    def test_2D_descriptors_multithread(self):
        """Test the dimensions of the output descriptor dataframes calculated by different processes."""
        values = self.cmp.calculate(MOLECULES_2D, show_banner=False, njobs=-1, chunksize=1)
        self.assertEqual(values.shape, (len(MOLECULES_2D), 632))
        self.assertEqual(len(values.columns.unique().tolist()), 632)

    def test_3D_descriptors(self):
        """Test the dimensions of the output descriptor dataframe."""
        values = self.cmp3d.calculate(MOLECULES_3D[:2], show_banner=False)
        self.assertEqual(values.shape, (len(MOLECULES_3D[:2]), 1184))
        self.assertEqual(len(values.columns.unique().tolist()), 1184)

    def test_3D_descriptors_multithread(self):
        """Test the dimensions of the output descriptor dataframes calculated by different processes."""
        values = self.cmp3d.calculate(MOLECULES_3D[:2], show_banner=False, njobs=-1, chunksize=1)
        self.assertEqual(values.shape, (len(MOLECULES_3D[:2]), 1184))
        self.assertEqual(len(values.columns.unique().tolist()), 1184)

    def test_2D_fingerprints_2D_descriptors(self):
        """Test the dimensions of the output descriptor dataframe."""
        values = self.cmpfp.calculate(MOLECULES_2D, show_banner=False)
        self.assertEqual(values.shape, (len(MOLECULES_2D), 14551))
        self.assertEqual(len(values.columns.unique().tolist()), 14551)

    def test_2D_fingerprints_2D_descriptors_multithread(self):
        """Test the dimensions of the output descriptor dataframes calculated by different processes."""
        values = self.cmpfp.calculate(MOLECULES_2D, show_banner=False, njobs=-1, chunksize=1)
        self.assertEqual(values.shape, (len(MOLECULES_2D), 14551))
        self.assertEqual(len(values.columns.unique().tolist()), 14551)

    def test_2D_fingerprints(self):
        for mol in MOLECULES_2D:
            values = Fingerprint.get_all_fps(mol)
            self.assertEqual(len(values), 13919)

    def test_3D_fingerprints(self):
        for mol in MOLECULES_3D:
            values = Fingerprint3D.get_all_fps(mol)
            self.assertEqual(len(values), 2048)

    def test_3D_fingerprints_2D_descriptors(self):
        """Test the dimensions of the output descriptor dataframe."""
        values = self.cmp3dfp.calculate(MOLECULES_3D[:2], show_banner=False)
        self.assertEqual(values.shape, (len(MOLECULES_3D[:2]), 17151))
        self.assertEqual(len(values.columns.unique().tolist()), 17151)

    def test_3D_fingerprints_2D_descriptors_multithread(self):
        """Test the dimensions of the output descriptor dataframes calculated by different processes."""
        values = self.cmp3dfp.calculate(MOLECULES_3D[:2], show_banner=False, njobs=-1, chunksize=1)
        self.assertEqual(values.shape, (len(MOLECULES_3D[:2]), 17151))
        self.assertEqual(len(values.columns.unique().tolist()), 17151)



# -*- coding: utf-8 -*-


"""Moreau-Broto autocorrelation descriptors."""


import numpy as np
from rdkit import Chem

from .atomproperty import get_relative_atomic_property


class MoreauBroto:
    """Moreau-Broto autocorrelation descriptors."""

    @staticmethod
    def _calculate_moreau_broto_autocorrelation(mol: Chem.Mol, lag: int = 1, propertylabel: str = 'm') -> float:
        """Calculate weighted Moreau-Broto autocorrelation descriptors.

        :param lag: topological distance between atom i and atom j.
        :param propertylabel: type of weighted property
        """
        Natom = mol.GetNumAtoms()
        distance_matrix = Chem.GetDistanceMatrix(mol)
        res = 0.0
        for i in range(Natom):
            for j in range(Natom):
                if distance_matrix[i, j] == lag:
                    atom1 = mol.GetAtomWithIdx(i)
                    atom2 = mol.GetAtomWithIdx(j)
                    temp1 = get_relative_atomic_property(element=atom1.GetSymbol(), propertyname=propertylabel)
                    temp2 = get_relative_atomic_property(element=atom2.GetSymbol(), propertyname=propertylabel)
                    res = res + temp1 * temp2
                else:
                    res = res + 0.0
        return np.log(res / 2 + 1)

    @staticmethod
    def calculate_moreau_broto_auto_mass(mol: Chem.Mol) -> dict:
        """Calculate Moreau-Broto autocorrelation with carbon-scaled atomic mass."""
        res = {}
        for i in range(8):
            res[f'ATSm{i + 1}'] = MoreauBroto._calculate_moreau_broto_autocorrelation(mol, lag=i + 1, propertylabel='m')
        return res

    @staticmethod
    def calculate_moreau_broto_auto_volume(mol: Chem.Mol) -> dict:
        """Calculate Moreau-Broto autocorrelation with carbon-scaled atomic van der Waals volume."""
        res = {}
        for i in range(8):
            res[f'ATSv{i + 1}'] = MoreauBroto._calculate_moreau_broto_autocorrelation(mol, lag=i + 1, propertylabel='V')
        return res

    @staticmethod
    def calculate_moreau_broto_auto_electronegativity(mol: Chem.Mol) -> dict:
        """Calculate Moreau-Broto autocorrelation with carbon-scaled atomic Sanderson electronegativity."""
        res = {}
        for i in range(8):
            res[f'ATSe{i + 1}'] = MoreauBroto._calculate_moreau_broto_autocorrelation(mol, lag=i + 1, propertylabel='En')
        return res

    @staticmethod
    def calculate_moreau_broto_auto_polarizability(mol: Chem.Mol) -> dict:
        """Calculate Moreau-Broto autocorrelation with carbon-scaled atomic polarizability."""
        res = {}
        for i in range(8):
            res[f'ATSp{i + 1}'] = MoreauBroto._calculate_moreau_broto_autocorrelation(mol, lag=i + 1, propertylabel='alapha')
        return res

    @staticmethod
    def get_all(mol: Chem.Mol) -> dict:
        """Calculate all (32) Moreau-Broto autocorrelation descriptors."""
        res = {}
        res.update(MoreauBroto.calculate_moreau_broto_auto_mass(mol))
        res.update(MoreauBroto.calculate_moreau_broto_auto_volume(mol))
        res.update(MoreauBroto.calculate_moreau_broto_auto_electronegativity(mol))
        res.update(MoreauBroto.calculate_moreau_broto_auto_polarizability(mol))
        return res



# -*- coding: utf-8 -*-


"""Utility functions for chemopy."""

import math
import os
from pathlib import Path
from typing import List

import numpy as np
from openbabel import openbabel, pybel
from rdkit import Chem


def get_r(n: int) -> List[float]:
    """Calcuate the parameters R of the RDF equation."""
    R = []
    for i in range(2, n + 2):
        R.append(float(i * 0.5))
    return R


def get_atom_distance(x: List[float], y: List[float]) -> float:
    """Calculate Euclidean distance between two atomic coordinates."""
    temp = [math.pow(x[0] - y[0], 2), math.pow(x[1] - y[1], 2), math.pow(x[2] - y[2], 2)]
    res = math.sqrt(sum(temp))
    return res


def get_geometrical_distance_matrix(CoordinateList: List[List[float]]) -> np.array:
    """Calculate distance matrix from coordinate list."""
    NAtom = len(CoordinateList)
    DistanceMatrix = np.zeros((NAtom, NAtom))
    for i in range(NAtom - 1):
        for j in range(i + 1, NAtom):
            DistanceMatrix[i, j] = get_atom_distance(CoordinateList[i], CoordinateList[j])
            DistanceMatrix[j, i] = DistanceMatrix[i, j]
    return DistanceMatrix


def is_in_subdirectory_tree(path: str, parent_path: str) -> bool:
    """Test if a path is in the subdirectory tree of another.

    :param path: the path to be tested
    :param parent_path: the possible parent directory
    """
    # Get full absolute paths
    path_ = os.path.realpath(path)
    parent_path_ = os.path.realpath(parent_path)
    # Get relative path
    relpath = os.path.relpath(parent_path_, path_)
    # Remove / and \ and '..'
    relpath = relpath.replace('/', '').replace('\\', '').replace('..', '')
    return len(relpath) == 0


def get_file_in_dir_from_ext(dirpath: str, ext: str) -> List[str]:
    """Identify the files in the current directory based on their extension.

    Does not support recursuve search of files.
    :param dirpath: path to directory
    :param ext: extensions (including .) to filter files on
    :return: list of files matching the required extension(s)
    """
    return [os.path.realpath(os.path.join(dirpath, x))
            for x in os.listdir(dirpath)
            if os.path.splitext(x)[1].lower() == ext.lower()]


def are_all_paths_absolute(paths: List[str]) -> bool:
    """Verify all paths are absolute."""
    return all(map(os.path.isabs, paths))


def are_all_paths_relative(paths: List[str]) -> bool:
    """Verify all paths are relative."""
    return all(map(lambda x: not os.path.isabs(x), paths))


def get_lastest_created_file(dirpath: str = None, filepaths: List[str] = None, strict: bool = True) -> str:
    """Get the file that was created last.

    Search can either be executed on an entire directory or a
    list of files that can come from  different directories.

    If only filepaths is provided, all paths must be absolute.
    If both dirpath and filepaths are provided, filepaths must not be absolute paths.

    :param dirpath: path to directory
    :param filepaths: paths of files
    :param strict: whether to stop when a file from filepaths does not exist
    """
    if (dirpath, filepaths) == (None, None):
        raise ValueError('Either dirpath or filepaths must be provided.')
    if dirpath is not None and filepaths is not None:
        if not are_all_paths_relative(filepaths):
            raise ValueError('filepaths must not be absolute when dirpath provided.')
        if not os.path.isdir(dirpath):
            raise NotADirectoryError(f'{dirpath} does not exist.')
        files = []
        for file_ in filepaths:
            full_path = os.path.realpath(os.path.join(dirpath, file_))
            if not os.path.isfile(full_path) and strict:
                raise FileNotFoundError(f'{full_path} does not exist.')
            elif os.path.isfile(full_path):
                files.append((full_path, Path(full_path).stat().st_ctime_ns))
        return sorted(files, key=lambda x: x[1], reverse=True)[0][0]
    elif dirpath is not None:
        if not os.path.isdir(dirpath):
            raise NotADirectoryError(f'{dirpath} does not exist.')
        files = []
        for file_ in os.listdir(dirpath):
            full_path = os.path.realpath(os.path.join(dirpath, file_))
            if os.path.isfile(full_path):
                files.append((full_path, Path(full_path).stat().st_ctime_ns))
        return sorted(files, key=lambda x: x[1], reverse=True)[0][0]
    else:
        if not are_all_paths_absolute(filepaths):
            raise ValueError('filepaths must be absolute when dirpath is not provided.')
        files = []
        for file_ in filepaths:
            if not os.path.isfile(file_) and strict:
                raise FileNotFoundError(f'{file_} does not exist.')
            elif os.path.isfile(file_):
                files.append((file_, Path(file_).stat().st_ctime_ns))
        return sorted(files, key=lambda x: x[1], reverse=True)[0][0]


def rdkit_to_openbabel_mol(mol: Chem.Mol) -> openbabel.OBMol:
    """Convert a RDKit molecule to an OpenBabel molecule.

    :param mol: RDKit molecule
    """
    obmol = openbabel.OBMol()
    # Add hydrogen atoms to complete molecule if needed
    rdkitmol = Chem.Mol(mol)
    # Perceive valence and ring information before assigning hydrogens
    rdkitmol.UpdatePropertyCache(strict=False)
    rdkitmol = Chem.AddHs(rdkitmol)
    # Kekulize molecule
    Chem.rdmolops.Kekulize(rdkitmol, clearAromaticFlags=True)
    # Add atoms
    for atom in mol.GetAtoms():
        # Create new atom and assign values
        obatom = obmol.NewAtom()
        obatom.SetAtomicNum(atom.GetAtomicNum())
        obatom.SetIsotope(atom.GetIsotope())
        obatom.SetFormalCharge(atom.GetFormalCharge())
        obatom.SetPartialCharge(atom.GetDoubleProp('_PartialCharge'))
        obatom.SetSpinMultiplicity(atom.GetNumRadicalElectrons() + 1)
    for bond in mol.GetBonds():
        obmol.AddBond(bond.GetBeginAtomIdx() + 1, bond.GetEndAtomIdx() + 1, int(bond.GetBondTypeAsDouble()))
    obmol.AssignSpinMultiplicity(True)
    return obmol


def rdkit_to_pybel_mol(mol: Chem.Mol) -> pybel.Molecule:
    """Convert a RDKit molecule to an OpenBabel.Pybel molecule.

    :param mol: RDKit molecule
    """
    return pybel.Molecule(rdkit_to_openbabel_mol(mol))


def openbabel_to_rdkit_mol(mol: openbabel.OBMol) -> Chem.Mol:
    """Convert an OpenBabel molecule to a RDKit molecule.

    :param mol: OpenBabel molecule
    """
    # Create an editable molecule
    rdkitmol = Chem.rdchem.EditableMol(Chem.rdchem.Mol())
    for obatom in openbabel.OBMolAtomIter(mol):
        # Create new atom and assign values
        atom = Chem.Atom(obatom.GetAtomicNum())
        atom.SetIsotope(obatom.GetIsotope())
        atom.SetFormalCharge(obatom.GetFormalCharge())
        atom.SetDoubleProp('_PartialCharge', obatom.GetPartialCharge())
        atom.SetNumRadicalElectrons(obatom.GetSpinMultiplicity() - 1 if obatom.GetSpinMultiplicity() != 0 else 0)
        # Add it to the current molecule
        rdkitmol.AddAtom(atom)

        orders = {1: Chem.rdchem.BondType.SINGLE,
                  2: Chem.rdchem.BondType.DOUBLE,
                  3: Chem.rdchem.BondType.TRIPLE,
                  4: Chem.rdchem.BondType.QUADRUPLE,
                  5: Chem.rdchem.BondType.QUINTUPLE,
                  1.5: Chem.rdchem.BondType.AROMATIC}
    for obbond in openbabel.OBMolBondIter(mol):
        rdkitmol.AddBond(obbond.GetBeginAtomIdx() - 1, obbond.GetEndAtomIdx() - 1, orders[obbond.GetBondOrder()])
    rdkitmol = rdkitmol.GetMol()
    Chem.SanitizeMol(rdkitmol)
    return rdkitmol


def pybel_to_rdkit_mol(mol: pybel.Molecule) -> Chem.Mol:
    """Convert an OpenBabel.Pybel molecule to a RDKit molecule.

    :param mol: OpenBabel.Pybel molecule
    """
    return openbabel_to_rdkit_mol(mol.OBMol)

def needs_hydrogens(mol: Chem.Mol) -> bool:
    """Return if the molecule lacks hydrogen atoms or not.

    :param mol: RDKit Molecule
    :return: True if the molecule lacks hydrogens.
    """
    for atom in mol.GetAtoms():
        nHNbrs = 0
        for nbr in atom.GetNeighbors():
            if nbr.GetAtomicNum() == 1:
                nHNbrs += 1
        noNeighbors = False
        if atom.GetTotalNumHs(noNeighbors) > nHNbrs:
            return True
    return False



# -*- coding: utf-8 -*-


"""3D Radial Distribution Function (RDF) descriptors."""

import math

import numpy as np
from rdkit import Chem

from .atomproperty import get_relative_atomic_property
from .geo_opt import read_coordinates
from .utils import get_geometrical_distance_matrix, get_r

_beta = 100

class RDF:
    """3D Radial Distribution Function descriptors."""

    @staticmethod
    def calculate_unweight_rdf(charge_coordinates):
        """Calculate unweighted RDF descriptors."""
        R = get_r(n=30)
        temp = []
    #    ChargeCoordinates=_ReadCoordinates('temp.arc')
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + math.exp(-_beta * math.pow(Ri - DM[j, k], 2))
            RDFresult[f'RDFU{kkk + 1}'] = res
        return RDFresult

    @staticmethod
    def calculate_charge_rdf(charge_coordinates):
        """Calculate RDF descriptors with Charge schemes."""
        R = get_r(n=30)
        temp = []
        Charge = []
    #    ChargeCoordinates=_ReadCoordinates('temp.arc')
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
            Charge.append(float(i[4]))
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + Charge[j] * Charge[k] * math.exp(-_beta * math.pow(Ri - DM[j, k], 2))
            RDFresult[f'RDFC{kkk + 1}'] = res
        return RDFresult

    @staticmethod
    def calculate_mass_rdf(mol: Chem.Mol, charge_coordinates):
        """Calculate RDF descriptors with Mass schemes."""
        mass = [atom.GetMass() for atom in Chem.AddHs(mol).GetAtoms()]
        R = get_r(n=30)
        temp = []
    #    ChargeCoordinates=_ReadCoordinates('temp.arc')
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + mass[j] * mass[k] * math.exp(-_beta * math.pow(Ri - DM[j, k], 2))
            RDFresult[f'RDFM{kkk + 1}'] = res / 144
        return RDFresult

    @staticmethod
    def calculate_polarizability_rdf(charge_coordinates):
        """Calculate RDF descriptors with Polarizability schemes."""
        R = get_r(n=30)
        temp = []
        polarizability = []
    #    ChargeCoordinates=_ReadCoordinates('temp.arc')
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
            polarizability.append(get_relative_atomic_property(i[0], 'alapha'))
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + polarizability[j] * polarizability[k] * math.exp(-_beta * math.pow(Ri - DM[j, k], 2))
            RDFresult[f'RDFP{kkk + 1}'] = res
        return RDFresult

    @staticmethod
    def calculate_sanderson_electronegativity_rdf(charge_coordinates):
        """Calculate RDF descriptors with Sanderson Electronegativity schemes."""
        R = get_r(n=30)
        temp = []
        EN = []
    #    ChargeCoordinates=_ReadCoordinates('temp.arc')
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
            EN.append(get_relative_atomic_property(i[0], 'En'))
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + EN[j] * EN[k] * math.exp(-_beta * math.pow(Ri - DM[j, k], 2))
            RDFresult[f'RDFE{kkk + 1}'] = res
        return RDFresult

    @staticmethod
    def calculate_vdw_volume_rdf(charge_coordinates):
        """Calculate RDF with atomic van der Waals volume shemes."""
        R = get_r(n=30)
        temp = []
        VDW = []
    #    ChargeCoordinates=_ReadCoordinates('temp.arc')
        for i in charge_coordinates:
            # if i[0]!='H':
            temp.append([float(i[1]), float(i[2]), float(i[3])])
            VDW.append(get_relative_atomic_property(i[0], 'V'))
        DM = get_geometrical_distance_matrix(temp)
        nAT = len(temp)
        RDFresult = {}
        for kkk, Ri in enumerate(R):
            res = 0.0
            for j in range(nAT - 1):
                for k in range(j + 1, nAT):
                    res = res + VDW[j] * VDW[k] * math.exp(-_beta * math.pow(Ri - DM[j, k], 2))
            RDFresult[f'RDFV{kkk + 1}'] = res
        return RDFresult

    @staticmethod
    def get_rdf_unweighed(arc_file):
        """Obtain all Unweighed RDF descriptors."""
        ChargeCoordinates = read_coordinates(arc_file)
        result = RDF.calculate_unweight_rdf(ChargeCoordinates)
        return result

    @staticmethod
    def get_rdf_charge(arc_file):
        """Obtain all RDF descriptors with Charge schemes."""
        ChargeCoordinates = read_coordinates(arc_file)
        result = RDF.calculate_charge_rdf(ChargeCoordinates)
        return result

    @staticmethod
    def get_rdf_mass(mol: Chem.Mol, arc_file):
        """Obtain all RDF descriptors with Mass schemes."""
        ChargeCoordinates = read_coordinates(arc_file)
        result = RDF.calculate_mass_rdf(mol, ChargeCoordinates)
        return result

    @staticmethod
    def get_rdf_polarizability(arc_file):
        """Obtain all RDF descriptors with Polarizability schemes."""
        ChargeCoordinates = read_coordinates(arc_file)
        result = RDF.calculate_polarizability_rdf(ChargeCoordinates)
        return result

    @staticmethod
    def get_rdf_sanderson_electronegativity(arc_file):
        """Obtain all RDF descriptors with Sanderson Electronegativity schemes."""
        ChargeCoordinates = read_coordinates(arc_file)
        result = RDF.calculate_sanderson_electronegativity_rdf(ChargeCoordinates)
        return result

    @staticmethod
    def get_rdf_vdw_volume(arc_file):
        """Obtain all RDF descriptors with VDW Volume schemes."""
        ChargeCoordinates = read_coordinates(arc_file)
        result = RDF.calculate_vdw_volume_rdf(ChargeCoordinates)
        return result

    @staticmethod
    def get_all(mol: Chem.Mol, arc_file: str) -> dict:
        """Obtain all (180) RDF descriptors with different (un)weighted schemes."""
        result = {}
        ChargeCoordinates = read_coordinates(arc_file)
        result.update(RDF.calculate_unweight_rdf(ChargeCoordinates))
        result.update(RDF.calculate_charge_rdf(ChargeCoordinates))
        result.update(RDF.calculate_mass_rdf(mol, ChargeCoordinates)) if mol is not None else np.nan
        result.update(RDF.calculate_polarizability_rdf(ChargeCoordinates))
        result.update(RDF.calculate_sanderson_electronegativity_rdf(ChargeCoordinates))
        result.update(RDF.calculate_vdw_volume_rdf(ChargeCoordinates))
        return result



# -*- coding: utf-8 -*-


"""Burden eigvenvalue descriptors."""

import numpy as np
import numpy.linalg
from rdkit import Chem
from natsort import natsorted

from .atomproperty import get_relative_atomic_property


class BCUT:
    """BCUT descriptors."""
    
    @staticmethod
    def _get_burden_matrix(mol: Chem.Mol, propertylabel: str = 'm') -> np.array:
        """Calculate weighted Burden matrix and eigenvalues."""
        mol = Chem.AddHs(mol)
        Natom = mol.GetNumAtoms()
        AdMatrix = Chem.GetAdjacencyMatrix(mol)
        bondindex = np.argwhere(AdMatrix)
        AdMatrix1 = np.array(AdMatrix, dtype=np.float32)
        # The diagonal elements of B, Bii, are either given by
        # the carbon normalized atomic mass,
        # van der Waals volume, Sanderson electronegativity,
        # and polarizability of atom i.
        for i in range(Natom):
            atom = mol.GetAtomWithIdx(i)
            temp = get_relative_atomic_property(element=atom.GetSymbol(), propertyname=propertylabel)
            AdMatrix1[i, i] = temp
        # The element of B connecting atoms i and j, Bij,
        # is equal to the square root of the bond
        # order between atoms i and j.
        for i in bondindex:
            bond = mol.GetBondBetweenAtoms(int(i[0]), int(i[1]))
            if bond.GetBondType().name == 'SINGLE':
                AdMatrix1[i[0], i[1]] = np.sqrt(1)
            if bond.GetBondType().name == "DOUBLE":
                AdMatrix1[i[0], i[1]] = np.sqrt(2)
            if bond.GetBondType().name == "TRIPLE":
                AdMatrix1[i[0], i[1]] = np.sqrt(3)
            if bond.GetBondType().name == "AROMATIC":
                AdMatrix1[i[0], i[1]] = np.sqrt(1.5)
        # All other elements of B (corresponding non-bonded
        # atom pairs) are set to 0.001
        bondnonindex = np.argwhere(AdMatrix == 0)
        for i in bondnonindex:
            if i[0] != i[1]:
                AdMatrix1[i[0], i[1]] = 0.001
        return np.real(np.linalg.eigvals(AdMatrix1))

    @staticmethod
    def calculate_burden_mass(mol: Chem.Mol) -> dict:
        """Calculate (16) Burden descriptors from atomic mass."""
        temp =  BCUT._get_burden_matrix(mol, propertylabel='m')
        temp1 = np.sort(temp[temp >= 0])
        temp2 = np.sort(np.abs(temp[temp < 0]))
        if len(temp1) < 8:
            temp1 = np.concatenate((np.zeros(8), temp1))
        if len(temp2) < 8:
            temp2 = np.concatenate((np.zeros(8), temp2))
        bcut = ["bcutm16", "bcutm15", "bcutm14", "bcutm13", "bcutm12", "bcutm11", "bcutm10",
                "bcutm9", "bcutm8", "bcutm7", "bcutm6", "bcutm5", "bcutm4", "bcutm3",
                "bcutm2", "bcutm1"]
        bcutvalue = np.concatenate((temp2[-8:], temp1[-8:]))
        bcutvalue = [i for i in bcutvalue]
        res = dict(natsorted(dict(zip(bcut, bcutvalue)).items()))
        return res

    @staticmethod
    def calculate_burden_vdw(mol: Chem.Mol) -> dict:
        """Calculate (16) Burden descriptors from atomic volumes."""
        temp = BCUT._get_burden_matrix(mol, propertylabel='V')
        temp1 = np.sort(temp[temp >= 0])
        temp2 = np.sort(np.abs(temp[temp < 0]))
        if len(temp1) < 8:
            temp1 = np.concatenate((np.zeros(8), temp1))
        if len(temp2) < 8:
            temp2 = np.concatenate((np.zeros(8), temp2))
        bcut = ["bcutv16", "bcutv15", "bcutv14", "bcutv13", "bcutv12", "bcutv11", "bcutv10",
                "bcutv9", "bcutv8", "bcutv7", "bcutv6", "bcutv5", "bcutv4", "bcutv3",
                "bcutv2", "bcutv1"]
        bcutvalue = np.concatenate((temp2[-8:], temp1[-8:]))
        bcutvalue = [i for i in bcutvalue]
        res = dict(natsorted(dict(zip(bcut, bcutvalue)).items()))
        return res

    @staticmethod
    def calculate_burden_electronegativity(mol: Chem.Mol) -> dict:
        """Calculate (16) Burden descriptors from atomic electronegativity."""
        temp = BCUT._get_burden_matrix(mol, propertylabel='En')
        temp1 = np.sort(temp[temp >= 0])
        temp2 = np.sort(np.abs(temp[temp < 0]))
        if len(temp1) < 8:
            temp1 = np.concatenate((np.zeros(8), temp1))
        if len(temp2) < 8:
            temp2 = np.concatenate((np.zeros(8), temp2))
        bcut = ["bcute16", "bcute15", "bcute14", "bcute13", "bcute12", "bcute11", "bcute10",
                "bcute9", "bcute8", "bcute7", "bcute6", "bcute5", "bcute4", "bcute3",
                "bcute2", "bcute1"]
        bcutvalue = np.concatenate((temp2[-8:], temp1[-8:]))
        bcutvalue = [i for i in bcutvalue]
        res = dict(natsorted(dict(zip(bcut, bcutvalue)).items()))
        return res

    @staticmethod
    def calculate_burden_polarizability(mol: Chem.Mol) -> dict:
        """Calculate (16) Burden descriptors from atomic polarizability."""
        temp = BCUT._get_burden_matrix(mol, propertylabel='alapha')
        temp1 = np.sort(temp[temp >= 0])
        temp2 = np.sort(np.abs(temp[temp < 0]))
        if len(temp1) < 8:
            temp1 = np.concatenate((np.zeros(8), temp1))
        if len(temp2) < 8:
            temp2 = np.concatenate((np.zeros(8), temp2))
        bcut = ["bcutp16", "bcutp15", "bcutp14", "bcutp13", "bcutp12", "bcutp11", "bcutp10",
                "bcutp9", "bcutp8", "bcutp7", "bcutp6", "bcutp5", "bcutp4", "bcutp3",
                "bcutp2", "bcutp1"]
        bcutvalue = np.concatenate((temp2[-8:], temp1[-8:]))
        bcutvalue = [i for i in bcutvalue]
        res = dict(natsorted(dict(zip(bcut, bcutvalue)).items()))
        return res

    @staticmethod
    def get_all(mol: Chem.Mol) -> dict:
        """Calculate all (64) Burden descriptors."""
        bcut = {}
        bcut.update(BCUT.calculate_burden_mass(mol))
        bcut.update(BCUT.calculate_burden_vdw(mol))
        bcut.update(BCUT.calculate_burden_electronegativity(mol))
        bcut.update(BCUT.calculate_burden_polarizability(mol))
        return bcut



# -*- coding: utf-8 -*-


"""Main classes for the calculation of molecular descriptors."""

import os
from copy import deepcopy
from typing import List, Optional

import more_itertools
import numpy as np
import pandas as pd
from bounded_pool_executor import BoundedProcessPoolExecutor
from rdkit import Chem

from . import getmol, geo_opt
from .basak import Basak
from .bcut import BCUT
from .charge import Charge
from .connectivity import Connectivity
from .constitution import Constitution
from .cpsa import CPSA
from .estate import EState
from .fingerprint import Fingerprint
from .fingerprint3d import Fingerprint3D
from .geary import Geary
from .geometric import Geometric
from .kappa import Kappa
from .moe import MOE
from .molproperty import MolecularProperties
from .moran import Moran
from .moreaubroto import MoreauBroto
from .morse import MoRSE
from .quanchem import QuantumChemistry
from .rdf import RDF
from .topology import Topology
from .whim import WHIM


class ChemoPy:
    """Molecular descriptor calculator."""

    def __init__(self, ignore_3D: bool = True, include_fps: bool = False):
        """Instantiate a molecular descriptor calculator.

        :param ignore_3D: Avoid calculating 3D molecular descriptors (default: True).
        If `False`, then molecules must have 3D conformers.
        :param include_fps: Should 2D fingerprints be also calculated (default: False).
        """
        self.ignore_3D = ignore_3D
        self.include_fps = include_fps

    def calculate(self, mols: List[Chem.Mol], show_banner: bool = True, njobs: int = 1,
                  chunksize: Optional[int] = 1000) -> pd.DataFrame:
        """Calculate molecular descriptors and/or fingerprints.

        :param mols: RDKit molecules for which descriptors/fingerprints should be calculated
        :param show_banner: If True, show notice on ChemoPy's usage
        :param njobs: number of concurrent processes
        :param chunksize: number of molecules to be processed by a process; ignored if njobs is 1
        :return: a pandas DataFrame containing all ChemoPy descriptor/fingerprint values
        """
        if show_banner:
            self._show_banner()
        if njobs < 0:
            njobs = os.cpu_count() - njobs + 1
        # Parallelize should need be
        if njobs > 1:
            with BoundedProcessPoolExecutor(max_workers=njobs) as worker:
                futures = [worker.submit(self._calculate, list(chunk))
                           for chunk in more_itertools.batched(mols, chunksize)
                           ]
            return pd.concat([future.result()
                              for future in futures]
                             ).reset_index(drop=True).fillna(0)
        # Single process
        return self._calculate(list(mols))

    def _show_banner(self):
        """Print info message for citing."""
        print("""ChemoPy computes 1178 descriptor values (632 2D and 546 3D)
of 16 different chemical feature groups along with 11 2D and 1 3D fingerprint.
It relies the semi-empirical quantum chemistry program MOPAC to compute all
of its 3D molecular descriptors and fingerprint.

###################################

Should you publish results based on the ChemoPy descriptors, please cite:

ChemoPy: freely available python package for computational biology and chemoinformatics
Dong-Sheng Cao, Qing-Song Xu, Qian-Nan Hu and Yi-Zeng Liang
Bioinformatics 2013, 29(8), 1092–1094
DOI: 10.1093/bioinformatics/btt105

###################################

""")

    def _calculate(self, mols: List[Chem.Mol]) -> pd.DataFrame:
        """Calculate ChemoPy descriptors/fingerprints on one process.

        :param mols: RDKit molecules for which chemoPy descriptors should be calculated
        :return: a pandas DataFrame containing all chemoPy descriptor values
        """
        descs_2D = [Constitution.get_all, Topology.get_all, Connectivity.get_all, Kappa.get_all, Basak.get_all,
                    EState.get_all_fps, EState.get_all_descriptors, BCUT.get_all, MoreauBroto.get_all,
                    Moran.get_all, Geary.get_all, Charge.get_all, MolecularProperties.get_all, MOE.get_all]
        descs_3D = []
        # Sanity check for 3D descriptors
        if not self.ignore_3D:
            for mol in mols:
                confs = list(mol.GetConformers())
                if not (len(confs) > 0 and confs[-1].Is3D()):
                    raise ValueError('Cannot calculate 3D descriptors for a conformer-less molecule')
            descs_3D.extend([Geometric.get_all, CPSA.get_all, WHIM.get_all,
                             MoRSE.get_all, RDF.get_all, QuantumChemistry.get_all])
        # Include fingerprints?
        if self.include_fps:
            descs_2D.append(Fingerprint.get_all_fps)
            # Include 3D fingerprint(s)
            if not self.ignore_3D:
                descs_3D.append(Fingerprint3D.calculate_e3fp)
        # Calculate descriptors
        all_values = []
        skipped_mols = []
        for i, mol in enumerate(mols):
            mol_values = {}
            # 2D descriptors
            for desc_2D in descs_2D:
                try:
                    tmp = desc_2D(mol)
                    mol_values.update(tmp)
                except Exception as e:
                    raise e
                    pass
            # 3D descriptors
            if len(descs_3D):
                try:
                    dir_, arc_file = geo_opt.get_arc_file(mol, verbose=False)
                    mol3d = geo_opt.get_optimized_mol(arc_file=arc_file)
                    for desc_3D in descs_3D:
                        try:
                            mol_values.update(desc_3D(mol=mol3d, arc_file=arc_file))
                        except Exception as e:
                            raise e
                            pass
                    # Dispose of MOPAC files
                    try:
                        dir_.close()
                    except Exception as e:
                        raise e
                        pass
                except Exception as e:
                    raise e
                    # Add molecule index to those skipped
                    skipped_mols.append(i)
            all_values.append(mol_values)
        # Transform to pandas dataframe
        results = pd.DataFrame(all_values)
        if results.empty:
            return results
        # Insert lines for skipped molecules
        if len(skipped_mols):
            results = pd.DataFrame(np.insert(results.values, skipped_mols,
                                             values=[np.NaN] * len(results.columns),
                                             axis=0),
                                   columns=results.columns)
        return results

    def _multiproc_calculate(self, mols: List[Chem.Mol]) -> pd.DataFrame:
        """Calculate ChemoPy descriptors/fingerprints in thread-safe manner.

        :param mols: RDKit molecules for which PaDEL descriptors should be calculated.
         Only the last conformer of molecules is considered.
        :return: a pandas DataFrame containing all PaDEL desciptor values and the path to the temp dir to be removed
        """
        # Copy self instance to make thread safe
        cmp = deepcopy(self)
        # Run copy
        result = cmp.calculate(mols, show_banner=False, njobs=1)
        return result

    @staticmethod
    def get_details(desc_name: Optional[str] = None) -> pd.DataFrame:
        """Obtain details about one/all descriptor(s)/fingerprint(s).

        :param desc_name: Name of descriptor/fingerprint (default: None).
        If `None`, obtain details for all.
        :return: The details about the descriptor(s)/fingerprint(s)
        """
        details = pd.read_json(os.path.abspath(os.path.join(__file__, os.pardir, 'descs.json')), orient='index')
        if desc_name is not None:
            if desc_name not in details.Name.tolist():
                raise ValueError(f'descriptor name {desc_name} is not available')
            details = details[details.Name == desc_name]
        return details



# -*- coding: utf-8 -*-


"""Molecule format parsers."""

import re
import urllib.request

from rdkit import Chem


class MolFrom:
    """Obtain molecular structures from common resources."""

    @staticmethod
    def CAS(cas_id: str = "") -> Chem.Mol:
        """Download molecule from ChemNet by CAS ID."""
        casid = cas_id.strip()
        request = urllib.request.Request(f'http://www.chemnet.com/cas/supplier.cgi?terms={casid}&l=&exact=dict')
        with urllib.request.urlopen(request) as response:
            temp = [line.decode("utf-8") for line in response.readlines()]
        for i in temp:
            if re.findall('InChI=', i) == ['InChI=']:
                k = i.split('    <td align="left">')
                kk = k[1].split('</td>\r\n')
                if kk[0][0:5] == "InChI":
                    res = kk[0]
                else:
                    res = "None"
        m = Chem.MolFromInchi(res.strip())
        return m

    @staticmethod
    def EBI(chebi_id: str = "") -> Chem.Mol:
        """Donwload molecule from ChEBI or ChEMBL using ChEBI or ChEMBL id."""
        chid = chebi_id.strip().upper()
        if chid.startswith('CHEBI'):
            request = urllib.request.Request(f'https://www.ebi.ac.uk/chebi/saveStructure.do?sdf=true&chebiId={chid}')
            with urllib.request.urlopen(request) as response:
                sdf = response.read()
            if not len(sdf):
                raise Exception(f'Not a valid ChEBI ID: {chid}')
        elif chid.startswith('CHEMBL'):
            request = urllib.request.Request(f'https://www.ebi.ac.uk/chembl/api/data/molecule?chembl_id={chid}&format=sdf')
            with urllib.request.urlopen(request) as response:
                sdf = response.read().decode()
            # xml_tree = ET.fromstring(xml)
            # structure = xml_tree.findall('./molecules/molecule/molecule_structures/molfile')
            if not len(sdf):
                raise Exception(f'Not a valid ChEMBL ID: {chid}')
            # sdf = structure[0].text
        else:
            raise Exception('Valid ID starts with CHEBI: or CHEMBL')
        m = Chem.MolFromMolBlock(sdf)
        return m

    @staticmethod
    def NCBI(cid: str = "") -> Chem.Mol:
        """Download molecule from PubChem using PubChem CID."""
        cid = cid.strip()
        cid = cid.upper().replace('CID', '') if 'CID' in cid.upper() else cid
        request = urllib.request.Request(f'http://pubchem.ncbi.nlm.nih.gov/summary/summary.cgi?cid={cid}&disopt=SaveSDF')
        with urllib.request.urlopen(request) as response:
            m = Chem.MolFromMolBlock(response.read())
        return m

    @staticmethod
    def DrugBank(drugbank_id: str = "") -> Chem.Mol:
        """Download molecule from DrugBank using DrugBank id."""
        dbid = drugbank_id.strip()
        request = urllib.request.Request(f'https://go.drugbank.com/structures/small_molecule_drugs/{dbid}.sdf',
                                         headers={"User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) "
                                                                "AppleWebKit/537.36 (KHTML, like Gecko) "
                                                                "Chrome/39.0.2171.95 "
                                                                "Safari/537.36"})
        with urllib.request.urlopen(request) as response:
            m = Chem.MolFromMolBlock(response.read())
        return m

    @staticmethod
    def KEGG(kegg_id: str = ""):
        """Download molecule from KEGG using KEGG id."""
        ID = str(kegg_id)
        request = urllib.request.Request(f'http://www.genome.jp/dbget-bin/www_bget?-f+m+drug+{ID}')
        with urllib.request.urlopen(request) as response:  # nosec: S310
            m = Chem.MolFromMolBlock(response.read())
        return m



# -*- coding: utf-8 -*-


"""Multiple 2D molecular fingerprints."""


import inspect
from typing import Optional

from map4 import MAP4Calculator
from mhfp.encoder import MHFPEncoder
from openbabel import pybel
from rdkit import Chem, DataStructs
from rdkit.Chem import MACCSkeys
from rdkit.Chem.rdMolDescriptors import (GetHashedAtomPairFingerprintAsBitVect,
                                         GetMorganFingerprintAsBitVect,
                                         GetHashedTopologicalTorsionFingerprintAsBitVect)

from .estate import EState

similaritymeasure = [i[0] for i in DataStructs.similarityFunctions]

class Fingerprint:
    """Molecular 2D fingerprints."""

    @staticmethod
    def calculate_rdk_fp(mol: Chem.Mol, nbits: int = 2048) -> dict:
        """Calculate a topological Daylight-like RDKit fingerprint.

        :param mols: the molecules
        :param nbits: Number of folded bits
        """
        bv = Chem.RDKFingerprint(mol, fpSize=nbits).ToList()
        values = dict(zip([f'RDK_{i+1}' for i in range(nbits)], bv))
        return values

    @staticmethod
    def calculate_maccs_fp(mol: Chem.Mol) -> dict:
        """Calculate the MACCS public keys (166 bits)."""
        bv = MACCSkeys.GenMACCSKeys(mol).ToList()
        values = dict(zip([f'MACCS_{i + 1}' for i in range(166)], bv))
        return values

    @staticmethod
    def calculate_fp2_fp(mol: Chem.Mol) -> dict:
        """Calculate topological FP2 fingerprints (1024 bits)."""
        m = pybel.readstring('smi', Chem.MolToSmiles(mol))
        on_bits = m.calcfp('FP2').bits
        values = dict(zip([f'FP2_{i + 1}' for i in range(1024)],
                          [1 if i + 1 in on_bits else 0 for i in range(1024)]))
        return values

    @staticmethod
    def calculate_fp3_fp(mol: Chem.Mol) -> dict:
        """Calculate FP3 fingerprints (55 bits)."""
        m = pybel.readstring('smi', Chem.MolToSmiles(mol))
        on_bits = m.calcfp('FP3').bits
        values = dict(zip([f'FP3_{i + 1}' for i in range(55)],
                          [1 if i + 1 in on_bits else 0 for i in range(55)]))
        return values

    @staticmethod
    def calculate_fp4_fp(mol: Chem.Mol) -> dict:
        """Calculate FP4 fingerprints (307 bits)."""
        m = pybel.readstring('smi', Chem.MolToSmiles(mol))
        on_bits = m.calcfp('FP4').bits
        values = dict(zip([f'FP4_{i + 1}' for i in range(307)],
                          [1 if i + 1 in on_bits else 0 for i in range(307)]))
        return values

    @staticmethod
    def calculate_estate_fp(mol: Chem.Mol) -> dict:
        """Calculate E-state fingerprints (79 bits)."""
        values = EState.calculate_estate_fingerprint(mol, implementation='chemopy', binary=True)
        return values

    @staticmethod
    def calculate_atompairs_fp(mol: Chem.Mol, nbits: int = 2048) -> dict:
        """Calculate atom pairs fingerprints.

        :param nbits: Number of folded bits
        """
        bv = GetHashedAtomPairFingerprintAsBitVect(mol, nBits=nbits).ToList()
        values = dict(zip([f'AtomPair_{i + 1}' for i in range(nbits)], bv))
        return values

    @staticmethod
    def calculate_topological_torsion_fp(mol: Chem.Mol, nbits: int = 2048) -> dict:
        """Calculate Topological Torsion fingerprints.

        :param nbits: Number of folded bits
        """
        bv = GetHashedTopologicalTorsionFingerprintAsBitVect(mol, nBits=nbits)
        values = dict(zip([f'TopolTorsions_{i + 1}' for i in range(nbits)], bv))
        return values

    @staticmethod
    def calculate_morgan_fp(mol: Chem.Mol, radius=2, nbits: int = 2048) -> dict:
        """Calculate Morgan fingerprints.

        :param radius: maximum radius of atom-centered substructures.
        :param rtype: Type of output, may either be:
                      bitstring (default), returns a binary string
                      rdkit, return the native rdkit DataStructs
                      dict, for a dict of bits turned on
        :param bits: Number of folded bits
        """
        bv = GetMorganFingerprintAsBitVect(mol, radius=radius, nBits=nbits)
        values = dict(zip([f'Morgan{radius * 2}_{i + 1}' for i in range(nbits)], bv))
        return values

    @staticmethod
    def calculate_secfp_fp(mol: Chem.Mol, radius: int = 3, nbits: int = 2048) -> dict:
        """Calculate the folded MinHash fingerpirnt (MHFP) of molecule.

        doi: 10.1186/s13321-018-0321-8.
        :param radius: maximum radius of atom-centered substructures.
        :param nbits: Number of folded bits
        """
        bv = MHFPEncoder.secfp_from_mol(mol, length=nbits, radius=radius,
                                        rings=True, kekulize=True, min_radius=1)
        values = dict(zip([f'SECFP{radius * 2}_{i + 1}' for i in range(nbits)], bv))
        return values

    @staticmethod
    def calculate_minhash_atompair_fp(mol: Chem.Mol, radius: int = 2, nbits: int = 2048) -> dict:
        """Calculate the MinHash fingerprint of Atom Pairs (MAP) of molecule.

        doi: 10.1186/s13321-020-00445-4.
        :param radius: maximum radius of atom-centered substructures.
        :param nbits: Number of folded bits (ignored if rtype != 'bitstring')
        """
        mapcalc = MAP4Calculator(radius=radius, dimensions=nbits, is_folded=True)
        bv = mapcalc.calculate(mol)
        values = dict(zip([f'MAP{radius * 2}_{i + 1}' for i in range(nbits)], bv))
        return values

    _fp_funcs = {'FP2': (calculate_fp2_fp, tuple()),
                 'FP3': (calculate_fp3_fp, tuple()),
                 'FP4': (calculate_fp4_fp, tuple()),
                 'MACCS': (calculate_maccs_fp, tuple()),
                 'Estate': (calculate_estate_fp, tuple()),
                 'topological': (calculate_rdk_fp, ('nbits')),
                 'atompairs': (calculate_atompairs_fp, ('nbits')),
                 'torsions': (calculate_topological_torsion_fp, ('nbits')),
                 'morgan': (calculate_morgan_fp, ('radius', 'nbits')),
                 'SECFP': (calculate_secfp_fp, ('radius', 'nbits')),
                 'MAP': (calculate_minhash_atompair_fp, ('radius', 'nbits'))
                 }

    @staticmethod
    def get_all_fps(mol: Chem.Mol, radius: Optional[int] = None, nbits: Optional[int] = None) -> dict:
        """Calculate all fingerprints."""
        values = {}
        for des_label, (func, supported_args) in Fingerprint._fp_funcs.items():
            if radius is not None and nbits is not None and 'radius' in supported_args and 'nbits' in supported_args:
                values.update(func(mol, radius=radius, nbits=nbits))
            elif radius is not None and 'radius' in supported_args:
                values.update(func(mol, radius=radius))
            elif nbits is not None and 'nbits' in supported_args:
                values.update(func(mol, nbits=nbits))
            else:
                values.update(func(mol))
        return values
